# !/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Forms."""
from netbox.forms import NetBoxModelForm
from netbox.plugins import get_plugin_config
from utilities.forms.rendering import FieldSet

from netbox_diode_plugin.models import Setting

__all__ = (
    "SettingsForm",
)


class SettingsForm(NetBoxModelForm):
    """Settings form."""

    fieldsets = (
        FieldSet(
            "diode_target",
        ),
    )

    class Meta:
        """Meta class."""

        model = Setting
        fields = ("diode_target",)

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        super().__init__(*args, **kwargs)

        diode_target_override = get_plugin_config(
            "netbox_diode_plugin", "diode_target_override"
        )

        if diode_target_override:
            self.fields["diode_target"].disabled = True
            self.fields["diode_target"].help_text = (
                "This field is not allowed to be modified."
            )



# !/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Models."""
from urllib.parse import urlparse

from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from netbox.models import NetBoxModel


def diode_target_validator(target):
    """Diode target validator."""
    try:
        parsed_target = urlparse(target)

        if parsed_target.scheme not in ["grpc", "grpcs"]:
            raise ValueError("target should start with grpc:// or grpcs://")
    except ValueError as exc:
        raise ValidationError(exc)


class Setting(NetBoxModel):
    """Setting model."""

    diode_target = models.CharField(max_length=255, validators=[diode_target_validator])

    class Meta:
        """Meta class."""

        verbose_name = "Settings"
        verbose_name_plural = "Settings"

    def __str__(self):
        """Return string representation."""
        return ""

    def get_absolute_url(self):
        """Return absolute URL."""
        return reverse("plugins:netbox_diode_plugin:settings")






#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API - Object resolution for diffing."""

import copy
import datetime
import json
import logging
import re
from collections import defaultdict
from functools import lru_cache
from uuid import uuid4

import graphlib
from django.utils.text import slugify
from extras.models.customfields import CustomField
from rest_framework import serializers

from .common import NON_FIELD_ERRORS, AutoSlug, ChangeSetException, UnresolvedReference, harmonize_formats, sort_ints_first
from .matcher import find_existing_object, fingerprints
from .plugin_utils import (
    CUSTOM_FIELD_OBJECT_REFERENCE_TYPE,
    apply_format_transformations,
    get_json_ref_info,
    get_primary_value,
    legal_fields,
)

logger = logging.getLogger("netbox.diode_data")

@lru_cache(maxsize=128)
def _camel_to_snake_case(name):
    """Convert camelCase string to snake_case."""
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()

# these are implied values pushed down to referenced objects.
_NESTED_CONTEXT = {
    "dcim.interface": {
        # interface.primary_mac_address -> mac_address.assigned_object = interface
        "primary_mac_address": lambda object_type, uuid: {
            "assigned_object_type": object_type,
            "assigned_object_id": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
    "virtualization.vminterface": {
        # interface.primary_mac_address -> mac_address.assigned_object = vinterface
        "primary_mac_address": lambda object_type, uuid: {
            "assigned_object_type": object_type,
            "assigned_object_id": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
    "virtualization.virtualmachine": {
        "primary_ip4": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
        "primary_ip6": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
    "dcim.virtualdevicecontext": {
        "primary_ip4": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
        "primary_ip6": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
}

def _no_context(object_type, uuid):
    return None

def _nested_context(object_type, uuid, field_name):
    return _NESTED_CONTEXT.get(object_type, {}).get(field_name, _no_context)(object_type, uuid)

_IS_CIRCULAR_REFERENCE = {
    "dcim.interface": frozenset(["primary_mac_address"]),
    "virtualization.vminterface": frozenset(["primary_mac_address"]),
    "dcim.device": frozenset(["primary_ip4", "primary_ip6", "oob_ip"]),
    "dcim.virtualdevicecontext": frozenset(["primary_ip4", "primary_ip6"]),
    "virtualization.virtualmachine": frozenset(["primary_ip4", "primary_ip6"]),
    "circuits.provider": frozenset(["accounts"]),
    "dcim.modulebay": frozenset(["module"]), # this isn't  allowed to be circular, but gives a better error
}

def _is_circular_reference(object_type, field_name):
    return field_name in _IS_CIRCULAR_REFERENCE.get(object_type, frozenset())

def transform_proto_json(proto_json: dict, object_type: str, supported_models: dict) -> list[dict]: # noqa: C901
    """
    Transform keys of proto json dict to flattened dictionaries with model field keys.

    This also handles placing `_type` fields for generic references,
    a certain form of deduplication and resolution of existing objects.
    """
    entities = _transform_proto_json_1(proto_json, object_type)

    entities = _topo_sort(entities)
    deduplicated = _fingerprint_dedupe(entities)
    deduplicated = _topo_sort(deduplicated)
    _set_auto_slugs(deduplicated, supported_models)
    _handle_cached_scope(deduplicated, supported_models)
    resolved = _resolve_existing_references(deduplicated)
    _strip_cached_scope(resolved)
    defaulted = _set_defaults(resolved, supported_models)

    # handle post-create steps
    output = _handle_post_creates(defaulted)

    _check_unresolved_refs(output)
    for entity in output:
        entity.pop('_refs', None)

    return output

def _transform_proto_json_1(proto_json: dict, object_type: str, context=None) -> list[dict]: # noqa: C901
    uuid = str(uuid4())
    node = {
        "_object_type": object_type,
        "_uuid": uuid,
        "_refs": set(),
    }

    # handle camelCase protoJSON if provided...
    proto_json = _ensure_snake_case(proto_json, object_type)
    apply_format_transformations(proto_json, object_type)

    # context pushed down from parent nodes
    if context is not None:
        for k, v in context.items():
            if not k.startswith("_"):
                node[k] = v
            if isinstance(v, UnresolvedReference):
                node['_refs'].add(v.uuid)

    nodes = [node]
    post_create = None

    # special handling for custom fields
    custom_fields = dict.pop(proto_json, "custom_fields", {})
    if custom_fields:
        custom_fields, custom_fields_refs, nested = _prepare_custom_fields(object_type, custom_fields)
        node['custom_fields'] = custom_fields
        node['_refs'].update(custom_fields_refs)
        nodes += nested

    for key, value in proto_json.items():
        ref_info = get_json_ref_info(object_type, key)
        if ref_info is None:
            node[key] = copy.deepcopy(value)
            continue

        nested_context = _nested_context(object_type, uuid, ref_info.field_name)
        field_name = ref_info.field_name
        is_circular = _is_circular_reference(object_type, field_name)

        if ref_info.is_generic:
            node[field_name + "_type"] = ref_info.object_type
            field_name = field_name + "_id"

        refs = []
        ref_value = None
        if isinstance(value, list):
            ref_value = []
            for item in value:
                nested = _transform_proto_json_1(item, ref_info.object_type, nested_context)
                nodes += nested
                ref_uuid = nested[0]['_uuid']
                ref_value.append(UnresolvedReference(
                    object_type=ref_info.object_type,
                    uuid=ref_uuid,
                ))
                refs.append(ref_uuid)
        else:
            nested = _transform_proto_json_1(value, ref_info.object_type, nested_context)
            nodes += nested
            ref_uuid = nested[0]['_uuid']
            ref_value = UnresolvedReference(
                object_type=ref_info.object_type,
                uuid=ref_uuid,
            )
            refs.append(ref_uuid)

        if is_circular:
            if post_create is None:
                post_create = {
                    "_uuid": str(uuid4()),
                    "_object_type": object_type,
                    "_refs": set(),
                    "_instance": node['_uuid'],
                    "_is_post_create": True,
                }
            post_create[field_name] = ref_value
            post_create['_refs'].update(refs)
            post_create['_refs'].add(node['_uuid'])
            continue

        node[field_name] = ref_value
        node['_refs'].update(refs)

    if post_create:
        nodes.append(post_create)

    return nodes

def _ensure_snake_case(proto_json: dict, object_type: str) -> dict:
    fields = legal_fields(object_type)
    out = {}
    for k, v in proto_json.items():
        if k in fields or get_json_ref_info(object_type, k):
            out[k] = v
            continue
        snake_key = _camel_to_snake_case(k)
        if snake_key in fields or get_json_ref_info(object_type, snake_key):
            out[snake_key] = v
        else:
            # error?
            sanitized_k = k.replace('\n', '').replace('\r', '')
            sanitized_snake_key = snake_key.replace('\n', '').replace('\r', '')
            sanitized_object_type = object_type.replace('\n', '').replace('\r', '')
            logger.warning(f"Unknown field {sanitized_k}/{sanitized_snake_key} is not legal for {sanitized_object_type}, skipping...")
    return out


def _topo_sort(entities: list[dict]) -> list[dict]:
    """Topologically sort entities by reference."""
    by_uuid = {e['_uuid']: e for e in entities}
    graph = defaultdict(set)
    for entity in entities:
        graph[entity['_uuid']] = entity['_refs'].copy()

    try:
        ts = graphlib.TopologicalSorter(graph)
        order = tuple(ts.static_order())
        return [by_uuid[uuid] for uuid in order]
    except graphlib.CycleError as e:
        # TODO the cycle error references the cycle here ...
        raise ChangeSetException(f"Circular reference in entities: {e}", errors={
            NON_FIELD_ERRORS: {
                NON_FIELD_ERRORS: "Unable to resolve circular reference in entities",
            }
        })


def _set_defaults(entities: list[dict], supported_models: dict):
    out = []
    for entity in entities:
        entity = copy.deepcopy(entity)
        model_fields = supported_models.get(entity['_object_type'])
        if model_fields is None:
            raise serializers.ValidationError({
                NON_FIELD_ERRORS: [f"Model for object type {entity['_object_type']} is not supported"]
            })

        auto_slug = entity.pop("_auto_slug", None)
        if entity.get("_instance"):
            out.append(entity)
            continue

        if auto_slug:
            if auto_slug.field_name not in entity:
                entity[auto_slug.field_name] = auto_slug.value

        legal = legal_fields(entity['_object_type'])
        for field_name, field_info in model_fields.get('fields', {}).items():
            if field_name not in legal:
                continue
            if entity.get(field_name) is None and field_info.get("default") is not None:
                default = field_info["default"]
                if callable(default):
                    default = default()
                entity[field_name] = default
        set_custom_field_defaults(entity, model_fields['model'])
        out.append(harmonize_formats(entity))
    return out

def _handle_cached_scope(entities: list[dict], supported_models: dict):
    by_type_id = {
        (entity['_object_type'], entity['_uuid']): entity
        for entity in entities
    }
    for entity in entities:
        model = supported_models.get(entity['_object_type'], {}).get("model")
        if _has_cached_scope(model):
            _handle_cached_scope_1(entity, by_type_id)

def _strip_cached_scope(entities: list[dict]):
    for entity in entities:
        entity.pop("_region", None)
        entity.pop("_site_group", None)
        entity.pop("_site", None)
        entity.pop("_location", None)

@lru_cache(maxsize=256)
def _has_cached_scope(model):
    return  hasattr(model, "cache_related_objects") and hasattr(model, "scope")

def _handle_cached_scope_1(entity: dict, by_type_id: dict):
    # these are some auto-set fields that cache scope information,
    # some indexes rely on them. Here we attempt to emulate that behavior
    # for the purpose of matching.  These generally only exist after save.
    scope_type = entity.get("scope_type")
    scope_id = entity.get("scope_id")

    if scope_type and scope_id:
        scope = by_type_id.get((scope_type, scope_id.uuid))
        if scope_type == "dcim.region":
            _cache_region_ref(entity, scope_id)
        elif scope_type == "dcim.sitegroup":
            _cache_site_group_ref(entity, scope_id)
        elif scope_type == "dcim.site":
            _cache_site_ref(entity, scope_id)
            _cache_region_ref(entity, scope.get("region"))
            _cache_site_group_ref(entity, scope.get("group"))
        elif scope_type == "dcim.location":
            _cache_location_ref(entity, scope_id)
            site_ref = scope.get("site")
            if site_ref is not None and isinstance(site_ref, UnresolvedReference):
                _cache_site_ref(entity, site_ref)
                site_obj = by_type_id.get((site_ref.object_type, site_ref.uuid))
                if site_obj is not None:
                    _cache_region_ref(entity, site_obj.get("region"))
                    _cache_site_group_ref(entity, site_obj.get("group"))

def _cache_region_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_region"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def _cache_site_group_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_site_group"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def _cache_site_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_site"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def _cache_location_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_location"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def set_custom_field_defaults(entity: dict, model):
    """Set default values for custom fields in an entity."""
    custom_fields = CustomField.objects.get_for_model(model)
    if custom_fields:
        custom_field_data = entity.get('custom_fields')
        if custom_field_data is None:
            custom_field_data = {}
            entity['custom_fields'] = custom_field_data
        for cf in custom_fields:
            if cf.name not in custom_field_data or custom_field_data[cf.name] is None:
                custom_field_data[cf.name] = cf.default

def _set_auto_slugs(entities: list[dict], supported_models: dict):
    for entity in entities:
        model_fields = supported_models.get(entity['_object_type'])
        if model_fields is None:
            raise serializers.ValidationError({
                NON_FIELD_ERRORS: [f"Model for object type {entity['_object_type']} is not supported"]
            })

        for field_name, field_info in model_fields.get('fields', {}).items():
            if field_info["type"] == "SlugField" and entity.get(field_name) is None:
                slug = _generate_slug(entity['_object_type'], entity)
                if slug is not None:
                    # this is provisionally set but will not be used
                    # if the entity is identified by other means...
                    entity['_auto_slug'] = AutoSlug(field_name=field_name, value=slug)

def _generate_slug(object_type, data):
    """Generate a slug for a model instance."""
    source_value = get_primary_value(data, object_type)
    if source_value is not None:
        return slugify(str(source_value))
    return None

def _fingerprint_dedupe(entities: list[dict]) -> list[dict]: # noqa: C901
    """
    Deduplicates/merges entities by fingerprint.

    *list must be in topo order by reference already*
    """
    by_uuid = {}
    by_fp = {}
    deduplicated = []
    new_refs = {} # uuid -> uuid

    for entity in entities:
        if entity.get('_is_post_create'):
            fp = entity['_uuid']
            existing_uuid = None
        else:
            _update_unresolved_refs(entity, new_refs)
            fps = fingerprints(entity, entity['_object_type'])
            for fp in fps:
                existing_uuid = by_fp.get(fp)
                if existing_uuid is not None:
                    break

        if existing_uuid is None:
            new_entity = copy.deepcopy(entity)
            _update_unresolved_refs(new_entity, new_refs)
            primary_uuid = new_entity['_uuid']
            for fp in fps:
                by_fp[fp] = primary_uuid
            by_uuid[primary_uuid] = new_entity
            deduplicated.append(primary_uuid)
        else:
            existing = by_uuid[existing_uuid]
            new_refs[entity['_uuid']] = existing['_uuid']
            merged = _merge_nodes(existing, entity)
            _update_unresolved_refs(merged, new_refs)
            for fp in fps:
                by_fp[fp] = existing_uuid
            by_uuid[existing_uuid] = merged
            deduplicated.append(existing_uuid)

    return [by_uuid[u] for u in deduplicated]

def _merge_nodes(a: dict, b: dict) -> dict:
    """
    Merges two nodes.

    If there are any conflicts, an error is raised.
    Ignores conflicts in fields that start with an underscore,
    preferring a's value.
    """
    merged = copy.deepcopy(a)
    merged['_refs'] = a['_refs'] | b['_refs']

    for k, v in b.items():
        if k.startswith("_"):
            continue
        if k in merged and merged[k] != v:
            ov = {
                ok: v for ok, v in a.items()
                if ok != k and not ok.startswith("_")
            }
            raise serializers.ValidationError({
                NON_FIELD_ERRORS: [
                    f"Conflicting values for '{k}' merging duplicate {a.get('_object_type')},"
                    f" `{merged[k]}` != `{v}` other values : {ov}"]
            })
        merged[k] = v
    return merged


def _update_unresolved_refs(entity, new_refs):
    if entity.get('_is_post_create'):
        instance_uuid = entity['_instance']
        entity['_instance'] = new_refs.get(instance_uuid, instance_uuid)

    entity['_refs'] = {new_refs.get(r,r) for r in entity['_refs']}
    _update_dict_refs(entity, new_refs)


def _update_dict_refs(data, new_refs):
    for k, v in data.items():
        if isinstance(v, UnresolvedReference) and v.uuid in new_refs:
            v.uuid = new_refs[v.uuid]
        elif isinstance(v, (list, tuple)):
            for item in v:
                if isinstance(item, UnresolvedReference) and item.uuid in new_refs:
                    item.uuid = new_refs[item.uuid]
        elif isinstance(v, dict):
            _update_dict_refs(v, new_refs)


def _resolve_existing_references(entities: list[dict]) -> list[dict]:
    seen = {}
    new_refs = {}
    resolved = []

    for data in entities:
        object_type = data['_object_type']
        data = copy.deepcopy(data)
        _update_resolved_refs(data, new_refs)

        if data.get('_is_post_create'):
            resolved.append(data)
            continue

        existing = find_existing_object(data, object_type)
        if existing is not None:
            fp = (object_type, existing.id)
            if fp in seen:
                logger.warning(f"objects resolved to the same existing id after deduplication: {seen[fp]} and {data}")
            else:
                seen[fp] = data
            data['id'] = existing.id
            data['_instance'] = existing
            new_refs[data['_uuid']] = existing.id
            resolved.append(data)
        else:
            data['id'] = UnresolvedReference(object_type, data['_uuid'])
            _update_resolved_refs(data, new_refs)
            resolved.append(data)
    return resolved

def _update_resolved_refs(data, new_refs):
    for k, v in list(data.items()):
        if isinstance(v, UnresolvedReference) and v.uuid in new_refs:
            data[k] = new_refs[v.uuid]
        elif isinstance(v, (list, tuple)):
            new_items = []
            has_refs = False
            for item in v:
                if isinstance(item, UnresolvedReference) and item.uuid in new_refs:
                    new_items.append(new_refs[item.uuid])
                    has_refs = True
                else:
                    new_items.append(item)
            if has_refs:
                data[k] = sort_ints_first(new_items)
        elif isinstance(v, dict):
            _update_resolved_refs(v, new_refs)

def cleanup_unresolved_references(data: dict) -> list[str]:
    """Find and stringify unresolved references in fields."""
    unresolved = set()
    for k, v in data.items():
        if isinstance(v, UnresolvedReference):
            if k != 'id':
                unresolved.add(k)
            data[k] = str(v)
        elif isinstance(v, (list, tuple)):
            items = []
            for item in v:
                if isinstance(item, UnresolvedReference):
                    unresolved.add(k)
                    items.append(str(item))
                else:
                    items.append(item)
            data[k] = items
        elif isinstance(v, dict):
            for uu in cleanup_unresolved_references(v):
                unresolved.add(f"{k}.{uu}")
    return sorted(unresolved)

def _handle_post_creates(entities: list[dict]) -> list[str]:
    """Merges any unnecessary post-create steps for existing objects."""
    by_uuid = {e['_uuid']: (i, e) for i, e in enumerate(entities)}
    out = []
    for entity in entities:
        is_post_create = entity.pop('_is_post_create', False)
        if not is_post_create:
            out.append(entity)
            continue

        instance = entity.get('_instance')
        prior_index, prior_entity = by_uuid[instance]

        # a post create can be merged whenever the entities it relies on
        # already exist (were resolved) or there are no dependencies between
        # the object being updated and the post-create.
        can_merge = all(
            by_uuid[r][1].get('_instance') is not None
            for r in entity['_refs']
        ) or sorted(by_uuid[r][0] for r in entity['_refs'])[-1] == prior_index

        if can_merge:
            prior_entity.update([x for x in entity.items() if not x[0].startswith('_')])
        else:
            entity['id'] = prior_entity['id']
            out.append(entity)

    return out

def _check_unresolved_refs(entities: list[dict]) -> list[str]:
    seen = set()
    for e in entities:
        seen.add((e['_object_type'], e['_uuid']))
        for k, v in e.items():
            if isinstance(v, UnresolvedReference):
                if (v.object_type, v.uuid) not in seen:
                    raise ChangeSetException(
                        f"Unresolved reference {v} in {e} does not refer to a prior created object (circular reference?)",
                        errors={
                            e['_object_type']: {
                                k: ["unable to resolve reference"],
                            }
                        }
                    )


def _prepare_custom_fields(object_type: str, custom_fields: dict) -> tuple[dict, set, list]: # noqa: C901
    """Prepare custom fields for transformation."""
    out = {}
    refs = set()
    nodes = []
    for key, value in custom_fields.items():
        keyname = key
        try:
            value_type, value = _pop_custom_field_type_and_value(value)
            if value_type in ("text", "longText", "decimal", "boolean", "datetime", "selection", "url", "multipleSelection"):
                out[key] = value
            elif value_type == "date":
                # truncate to YYYY-MM-DD
                try:
                    out[key] = datetime.datetime.fromisoformat(value).strftime("%Y-%m-%d")
                except Exception:
                    out[key] = value
            elif value_type == "integer":
                out[key] = int(value)
            elif value_type == "json":
                out[key] = _prepare_custom_json(value)
            elif value_type == "object":
                nested = _prepare_custom_ref(value)
                ref = nested[0]
                refs.add(ref['_uuid'])
                nodes += nested
                out[key] = UnresolvedReference(
                    object_type=ref['_object_type'],
                    uuid=ref['_uuid'],
                )
            elif value_type == "multipleObjects":
                vals = []
                for i, item in enumerate(value):
                    keyname = f"{key}[{i}]"
                    nested = _prepare_custom_ref(value)
                    ref = nested[0]
                    refs.add(ref['_uuid'])
                    nodes += nested
                    vals.append(UnresolvedReference(
                        object_type=ref['_object_type'],
                        uuid=ref['_uuid'],
                    ))
                out[key] = vals
            else:
                raise serializers.ValidationError({
                    keyname: [f"Custom field {keyname} has unknown type: {value_type}"]
                })
        except ValueError as e:
            raise ChangeSetException(
                f"Custom field {keyname} is invalid: {value}",
                errors={
                    object_type: {keyname: [str(e)]},
                }
            )
    return out, refs, nodes


def _prepare_custom_json(data: dict) -> dict:
    try:
        return json.loads(data)
    except json.JSONDecodeError:
        raise ValueError("failed to parse as JSON")


def _pop_custom_field_type_and_value(data: dict):
    if not isinstance(data, dict) or len(data) != 1:
        raise ValueError("custom field value must be a dictionary with a single key")
    value_type, value = data.popitem()
    return value_type, value


def _prepare_custom_ref(data: dict) -> list[dict]:
    if not isinstance(data, dict) or len(data) != 1:
        raise ValueError("must be a dictionary with a single key")

    field_name, value = data.popitem()
    if not isinstance(value, dict):
        raise ValueError(f"{field_name} must be a dictionary")
    ref_info = get_json_ref_info(CUSTOM_FIELD_OBJECT_REFERENCE_TYPE, field_name)
    if ref_info is None:
        raise ValueError(f"{field_name} is not a supported custom field reference type")

    object_type = ref_info.object_type
    return _transform_proto_json_1(value, object_type)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs Inc
"""NetBox Diode Data - API supported models."""

import importlib
import logging
import time
from functools import lru_cache
from typing import List, Type

from django.apps import apps
from django.db import models
from django.db.models import ManyToOneRel
from django.db.models.fields import NOT_PROVIDED
from rest_framework import serializers
from utilities.api import get_serializer_for_model as netbox_get_serializer_for_model

logger = logging.getLogger(__name__)

# Supported apps
SUPPORTED_APPS = [
    "circuits",
    "dcim",
    "extras",
    "ipam",
    "virtualization",
    "vpn",
    "wireless",
    "tenancy",
]

# Models that are not supported
EXCLUDED_MODELS = [
    "TaggedItem",
    "Subscription",
    "ScriptModule",
    "Dashboard",
    "Notification",
]


def extract_supported_models() -> dict[str, dict]:
    """Extract supported models from NetBox."""
    supported_models = discover_models(SUPPORTED_APPS)

    logger.debug(f"Supported models: {supported_models}")

    models_to_process = supported_models
    extracted_models: dict[str, dict] = {}

    start_ts = time.time()
    while models_to_process:
        model = models_to_process.pop()
        try:
            fields, related_models = get_model_fields(model)
            if not fields:
                continue

            prerequisites = get_prerequisites(model, fields)
            object_type = f"{model._meta.app_label}.{model._meta.model_name}"
            extracted_models[object_type] = {
                "fields": fields,
                "prerequisites": prerequisites,
                "model": model,
            }
            for related_model in related_models:
                related_object_type = f"{related_model._meta.app_label}.{related_model._meta.model_name}"
                if (
                    related_object_type not in extracted_models
                    and related_object_type not in models_to_process
                ):
                    models_to_process.append(related_model)
        except Exception as e:
            logger.error(f"extract_supported_models: {model.__name__} error: {e}")

    finish_ts = time.time()
    lapsed_millis = (finish_ts - start_ts) * 1000
    logger.info(
        f"done extracting supported models in {lapsed_millis:.2f} milliseconds - extracted_models: {len(extracted_models)}"
    )

    return extracted_models


def get_prerequisites(model_class, fields) -> List[dict[str, str]]:
    """Get the prerequisite models for the model."""
    prerequisites: List[dict[str, str]] = []
    prerequisite_models = getattr(model_class, "prerequisite_models", [])

    for prereq in prerequisite_models:
        prereq_model = apps.get_model(prereq)

        for field_name, field_info in fields.items():
            related_model = field_info.get("related_model")
            prerequisite_info = {
                "field_name": field_name,
                "prerequisite_model": prereq_model,
            }
            if (
                prerequisite_info not in prerequisites
                and related_model
                and related_model.get("model_class_name") == prereq_model.__name__
            ):
                prerequisites.append(prerequisite_info)
                break

    return prerequisites


@lru_cache(maxsize=128)
def get_model_fields(model_class) -> tuple[dict, list]:
    """Get the fields for the model ordered as they are in the serializer."""
    related_models_to_process = []

    # Skip unsupported apps and excluded models
    if (
        model_class._meta.app_label not in SUPPORTED_APPS
        or model_class.__name__ in EXCLUDED_MODELS
    ):
        return {}, []

    try:
        # Get serializer fields to maintain order
        serializer_class = get_serializer_for_model(model_class)
        serializer_fields = serializer_class().get_fields()
        serializer_fields_names = list(serializer_fields.keys())
    except Exception as e:
        logger.error(f"Error getting serializer fields for model {model_class}: {e}")
        return {}, []

    # Get all model fields
    model_fields = {
        field.name: field
        for field in model_class._meta.get_fields()
        if field.__class__.__name__ not in ["CounterCacheField", "GenericRelation"]
    }

    # Reorder fields to match serializer order
    ordered_fields = {
        field_name: model_fields[field_name]
        for field_name in serializer_fields_names
        if field_name in model_fields
    }

    # Add remaining fields
    ordered_fields.update(
        {
            field_name: field
            for field_name, field in model_fields.items()
            if field_name not in ordered_fields
        }
    )

    fields_info = {}

    for field_name, field in ordered_fields.items():
        field_info = {
            "type": field.get_internal_type(),
            "required": not field.null and not field.blank,
            "is_many_to_one_rel": isinstance(field, ManyToOneRel),
            "is_numeric": field.get_internal_type()
            in [
                "IntegerField",
                "FloatField",
                "DecimalField",
                "PositiveIntegerField",
                "PositiveSmallIntegerField",
                "SmallIntegerField",
                "BigIntegerField",
            ],
        }

        # Handle default values
        default_value = None
        if hasattr(field, "default"):
            default_value = (
                field.default if field.default not in (NOT_PROVIDED, dict) else None
            )
        field_info["default"] = default_value

        # Handle related fields
        if field.is_relation:
            related_model = field.related_model
            if related_model:
                related_model_key = (
                    f"{related_model._meta.app_label}.{related_model._meta.model_name}"
                )
                related_model_info = {
                    "app_label": related_model._meta.app_label,
                    "model_name": related_model._meta.model_name,
                    "model_class_name": related_model.__name__,
                    "object_type": related_model_key,
                    "filters": get_field_filters(model_class, field_name),
                }
                field_info["related_model"] = related_model_info
                if (
                    related_model.__name__ not in EXCLUDED_MODELS
                    and related_model not in related_models_to_process
                ):
                    related_models_to_process.append(related_model)

        fields_info[field_name] = field_info

    return fields_info, related_models_to_process


@lru_cache(maxsize=128)
def get_field_filters(model_class, field_name):
    """Get filters for a field."""
    if hasattr(model_class, "_netbox_private"):
        return None

    try:
        filterset_name = f"{model_class.__name__}FilterSet"
        filterset_module = importlib.import_module(
            f"{model_class._meta.app_label}.filtersets"
        )
        filterset_class = getattr(filterset_module, filterset_name)

        _filters = set()
        field_filters = []
        for filter_name, filter_instance in filterset_class.get_filters().items():
            filter_by = getattr(filter_instance, "field_name", None)
            filter_field_extra = getattr(filter_instance, "extra", None)

            if not filter_name.startswith(field_name) or filter_by.endswith("_id"):
                continue

            if filter_by and filter_by not in _filters:
                _filters.add(filter_by)
                field_filters.append(
                    {
                        "filter_by": filter_by,
                        "filter_to_field_name": (
                            filter_field_extra.get("to_field_name", None)
                            if filter_field_extra
                            else None
                        ),
                    }
                )
        return list(field_filters) if field_filters else None
    except Exception as e:
        logger.error(
            f"Error getting field filters for model {model_class.__name__} and field {field_name}: {e}"
        )
        return None


@lru_cache(maxsize=128)
def get_serializer_for_model(model, prefix=""):
    """Cached wrapper for NetBox's get_serializer_for_model function."""
    return netbox_get_serializer_for_model(model, prefix)


def discover_models(root_packages: List[str]) -> list[Type[models.Model]]:
    """Discovers all model classes in specified root packages."""
    discovered_models = []

    # Look through all modules that might contain serializers
    module_names = [
        "api.serializers",
    ]

    for root_package in root_packages:
        logger.debug(f"Searching in root package: {root_package}")

        for module_name in module_names:
            full_module_path = f"{root_package}.{module_name}"
            try:
                module = __import__(full_module_path, fromlist=["*"])
            except ImportError:
                logger.error(f"Could not import {full_module_path}")
                continue

            # Find all serializer classes in the module
            for serializer_name in dir(module):
                serializer = getattr(module, serializer_name)
                if (
                    isinstance(serializer, type)
                    and issubclass(serializer, serializers.Serializer)
                    and serializer != serializers.Serializer
                    and serializer != serializers.ModelSerializer
                    and hasattr(serializer, "Meta")
                    and hasattr(serializer.Meta, "model")
                ):
                    model = serializer.Meta.model
                    if model not in discovered_models:
                        discovered_models.append(model)
                    logger.debug(
                        f"Discovered model: {model.__module__}.{model.__name__}"
                    )

    return discovered_models



# !/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Plugin Settings."""

from urllib.parse import urlparse

from django.contrib.auth import get_user_model
from netbox.plugins import get_plugin_config

__all__ = (
    "get_diode_auth_introspect_url",
    "get_diode_user",
)

User = get_user_model()


def _parse_diode_target(target: str) -> tuple[str, str, bool]:
    """Parse the target into authority, path and tls_verify."""
    parsed_target = urlparse(target)

    if parsed_target.scheme not in ["grpc", "grpcs"]:
        raise ValueError("target should start with grpc:// or grpcs://")

    tls_verify = parsed_target.scheme == "grpcs"

    authority = parsed_target.netloc

    return authority, parsed_target.path, tls_verify


def get_diode_auth_introspect_url():
    """Returns the Diode Auth introspect URL."""
    diode_target = get_plugin_config("netbox_diode_plugin", "diode_target")
    diode_target_override = get_plugin_config(
        "netbox_diode_plugin", "diode_target_override"
    )

    authority, path, tls_verify = _parse_diode_target(
        diode_target_override or diode_target
    )
    scheme = "https" if tls_verify else "http"
    path = path.rstrip("/")

    return f"{scheme}://{authority}{path}/auth/introspect"


def get_diode_user():
    """Returns the Diode user."""
    diode_username = get_plugin_config("netbox_diode_plugin", "diode_username")

    try:
        diode_user = User.objects.get(username=diode_username)
    except User.DoesNotExist:
        diode_user = User.objects.create(username=diode_username, is_active=True)

    return diode_user



"""Diode plugin helpers."""

# Generated code. DO NOT EDIT.
# Timestamp: 2025-04-13 16:50:25Z

import datetime
import decimal
import logging
from dataclasses import dataclass
from functools import lru_cache
from typing import Type

import netaddr
from core.models import ObjectType as NetBoxType
from django.contrib.contenttypes.models import ContentType
from django.db import models
from rest_framework.exceptions import ValidationError

logger = logging.getLogger(__name__)

@lru_cache(maxsize=256)
def get_object_type_model(object_type: str) -> Type[models.Model]:
    """Get the model class for a given object type."""
    app_label, model_name = object_type.split('.')
    object_content_type = NetBoxType.objects.get_by_natural_key(app_label, model_name)
    return object_content_type.model_class()

@lru_cache(maxsize=256)
def get_object_type(model_class: Type[models.Model]) -> str:
    """Get the object type for a given model class."""
    content_type = ContentType.objects.get_for_model(model_class)
    return content_type.app_label + '.' + content_type.model

@lru_cache(maxsize=256)
def content_type_id(object_type: str) -> int:
    """Get the content type id for a given object type."""
    app_label, model_name = object_type.split('.')
    object_content_type = NetBoxType.objects.get_by_natural_key(app_label, model_name)
    return object_content_type.id

@dataclass
class RefInfo:
    object_type: str
    field_name: str
    is_generic: bool = False
    is_many: bool = False

CUSTOM_FIELD_OBJECT_REFERENCE_TYPE = 'diode.custom_field_object_reference'

_JSON_REF_INFO = {
    'diode.custom_field_object_reference': {
        'asn': RefInfo(object_type='ipam.asn', field_name='object', is_generic=True),
        'asn_range': RefInfo(object_type='ipam.asnrange', field_name='object', is_generic=True),
        'aggregate': RefInfo(object_type='ipam.aggregate', field_name='object', is_generic=True),
        'cable': RefInfo(object_type='dcim.cable', field_name='object', is_generic=True),
        'cable_path': RefInfo(object_type='dcim.cablepath', field_name='object', is_generic=True),
        'cable_termination': RefInfo(object_type='dcim.cabletermination', field_name='object', is_generic=True),
        'circuit': RefInfo(object_type='circuits.circuit', field_name='object', is_generic=True),
        'circuit_group': RefInfo(object_type='circuits.circuitgroup', field_name='object', is_generic=True),
        'circuit_group_assignment': RefInfo(object_type='circuits.circuitgroupassignment', field_name='object', is_generic=True),
        'circuit_termination': RefInfo(object_type='circuits.circuittermination', field_name='object', is_generic=True),
        'circuit_type': RefInfo(object_type='circuits.circuittype', field_name='object', is_generic=True),
        'cluster': RefInfo(object_type='virtualization.cluster', field_name='object', is_generic=True),
        'cluster_group': RefInfo(object_type='virtualization.clustergroup', field_name='object', is_generic=True),
        'cluster_type': RefInfo(object_type='virtualization.clustertype', field_name='object', is_generic=True),
        'console_port': RefInfo(object_type='dcim.consoleport', field_name='object', is_generic=True),
        'console_server_port': RefInfo(object_type='dcim.consoleserverport', field_name='object', is_generic=True),
        'contact': RefInfo(object_type='tenancy.contact', field_name='object', is_generic=True),
        'contact_assignment': RefInfo(object_type='tenancy.contactassignment', field_name='object', is_generic=True),
        'contact_group': RefInfo(object_type='tenancy.contactgroup', field_name='object', is_generic=True),
        'contact_role': RefInfo(object_type='tenancy.contactrole', field_name='object', is_generic=True),
        'device': RefInfo(object_type='dcim.device', field_name='object', is_generic=True),
        'device_bay': RefInfo(object_type='dcim.devicebay', field_name='object', is_generic=True),
        'device_role': RefInfo(object_type='dcim.devicerole', field_name='object', is_generic=True),
        'device_type': RefInfo(object_type='dcim.devicetype', field_name='object', is_generic=True),
        'fhrp_group': RefInfo(object_type='ipam.fhrpgroup', field_name='object', is_generic=True),
        'fhrp_group_assignment': RefInfo(object_type='ipam.fhrpgroupassignment', field_name='object', is_generic=True),
        'front_port': RefInfo(object_type='dcim.frontport', field_name='object', is_generic=True),
        'ike_policy': RefInfo(object_type='vpn.ikepolicy', field_name='object', is_generic=True),
        'ike_proposal': RefInfo(object_type='vpn.ikeproposal', field_name='object', is_generic=True),
        'ip_address': RefInfo(object_type='ipam.ipaddress', field_name='object', is_generic=True),
        'ip_range': RefInfo(object_type='ipam.iprange', field_name='object', is_generic=True),
        'ip_sec_policy': RefInfo(object_type='vpn.ipsecpolicy', field_name='object', is_generic=True),
        'ip_sec_profile': RefInfo(object_type='vpn.ipsecprofile', field_name='object', is_generic=True),
        'ip_sec_proposal': RefInfo(object_type='vpn.ipsecproposal', field_name='object', is_generic=True),
        'interface': RefInfo(object_type='dcim.interface', field_name='object', is_generic=True),
        'inventory_item': RefInfo(object_type='dcim.inventoryitem', field_name='object', is_generic=True),
        'inventory_item_role': RefInfo(object_type='dcim.inventoryitemrole', field_name='object', is_generic=True),
        'l2vpn': RefInfo(object_type='vpn.l2vpn', field_name='object', is_generic=True),
        'l2vpn_termination': RefInfo(object_type='vpn.l2vpntermination', field_name='object', is_generic=True),
        'location': RefInfo(object_type='dcim.location', field_name='object', is_generic=True),
        'mac_address': RefInfo(object_type='dcim.macaddress', field_name='object', is_generic=True),
        'manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='object', is_generic=True),
        'module': RefInfo(object_type='dcim.module', field_name='object', is_generic=True),
        'module_bay': RefInfo(object_type='dcim.modulebay', field_name='object', is_generic=True),
        'module_type': RefInfo(object_type='dcim.moduletype', field_name='object', is_generic=True),
        'platform': RefInfo(object_type='dcim.platform', field_name='object', is_generic=True),
        'power_feed': RefInfo(object_type='dcim.powerfeed', field_name='object', is_generic=True),
        'power_outlet': RefInfo(object_type='dcim.poweroutlet', field_name='object', is_generic=True),
        'power_panel': RefInfo(object_type='dcim.powerpanel', field_name='object', is_generic=True),
        'power_port': RefInfo(object_type='dcim.powerport', field_name='object', is_generic=True),
        'prefix': RefInfo(object_type='ipam.prefix', field_name='object', is_generic=True),
        'provider': RefInfo(object_type='circuits.provider', field_name='object', is_generic=True),
        'provider_account': RefInfo(object_type='circuits.provideraccount', field_name='object', is_generic=True),
        'provider_network': RefInfo(object_type='circuits.providernetwork', field_name='object', is_generic=True),
        'rir': RefInfo(object_type='ipam.rir', field_name='object', is_generic=True),
        'rack': RefInfo(object_type='dcim.rack', field_name='object', is_generic=True),
        'rack_reservation': RefInfo(object_type='dcim.rackreservation', field_name='object', is_generic=True),
        'rack_role': RefInfo(object_type='dcim.rackrole', field_name='object', is_generic=True),
        'rack_type': RefInfo(object_type='dcim.racktype', field_name='object', is_generic=True),
        'rear_port': RefInfo(object_type='dcim.rearport', field_name='object', is_generic=True),
        'region': RefInfo(object_type='dcim.region', field_name='object', is_generic=True),
        'role': RefInfo(object_type='ipam.role', field_name='object', is_generic=True),
        'route_target': RefInfo(object_type='ipam.routetarget', field_name='object', is_generic=True),
        'service': RefInfo(object_type='ipam.service', field_name='object', is_generic=True),
        'site': RefInfo(object_type='dcim.site', field_name='object', is_generic=True),
        'site_group': RefInfo(object_type='dcim.sitegroup', field_name='object', is_generic=True),
        'tag': RefInfo(object_type='extras.tag', field_name='object', is_generic=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='object', is_generic=True),
        'tenant_group': RefInfo(object_type='tenancy.tenantgroup', field_name='object', is_generic=True),
        'tunnel': RefInfo(object_type='vpn.tunnel', field_name='object', is_generic=True),
        'tunnel_group': RefInfo(object_type='vpn.tunnelgroup', field_name='object', is_generic=True),
        'tunnel_termination': RefInfo(object_type='vpn.tunneltermination', field_name='object', is_generic=True),
        'vlan': RefInfo(object_type='ipam.vlan', field_name='object', is_generic=True),
        'vlan_group': RefInfo(object_type='ipam.vlangroup', field_name='object', is_generic=True),
        'vlan_translation_policy': RefInfo(object_type='ipam.vlantranslationpolicy', field_name='object', is_generic=True),
        'vlan_translation_rule': RefInfo(object_type='ipam.vlantranslationrule', field_name='object', is_generic=True),
        'vm_interface': RefInfo(object_type='virtualization.vminterface', field_name='object', is_generic=True),
        'vrf': RefInfo(object_type='ipam.vrf', field_name='object', is_generic=True),
        'virtual_chassis': RefInfo(object_type='dcim.virtualchassis', field_name='object', is_generic=True),
        'virtual_circuit': RefInfo(object_type='circuits.virtualcircuit', field_name='object', is_generic=True),
        'virtual_circuit_termination': RefInfo(object_type='circuits.virtualcircuittermination', field_name='object', is_generic=True),
        'virtual_circuit_type': RefInfo(object_type='circuits.virtualcircuittype', field_name='object', is_generic=True),
        'virtual_device_context': RefInfo(object_type='dcim.virtualdevicecontext', field_name='object', is_generic=True),
        'virtual_disk': RefInfo(object_type='virtualization.virtualdisk', field_name='object', is_generic=True),
        'virtual_machine': RefInfo(object_type='virtualization.virtualmachine', field_name='object', is_generic=True),
        'wireless_lan': RefInfo(object_type='wireless.wirelesslan', field_name='object', is_generic=True),
        'wireless_lan_group': RefInfo(object_type='wireless.wirelesslangroup', field_name='object', is_generic=True),
        'wireless_link': RefInfo(object_type='wireless.wirelesslink', field_name='object', is_generic=True),
    },
    'circuits.circuit': {
        'assignments': RefInfo(object_type='circuits.circuitgroupassignment', field_name='assignments', is_many=True),
        'provider': RefInfo(object_type='circuits.provider', field_name='provider'),
        'provider_account': RefInfo(object_type='circuits.provideraccount', field_name='provider_account'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'type': RefInfo(object_type='circuits.circuittype', field_name='type'),
    },
    'circuits.circuitgroup': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'circuits.circuitgroupassignment': {
        'group': RefInfo(object_type='circuits.circuitgroup', field_name='group'),
        'member_circuit': RefInfo(object_type='circuits.circuit', field_name='member', is_generic=True),
        'member_virtual_circuit': RefInfo(object_type='circuits.virtualcircuit', field_name='member', is_generic=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'circuits.circuittermination': {
        'circuit': RefInfo(object_type='circuits.circuit', field_name='circuit'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'termination_location': RefInfo(object_type='dcim.location', field_name='termination', is_generic=True),
        'termination_provider_network': RefInfo(object_type='circuits.providernetwork', field_name='termination', is_generic=True),
        'termination_region': RefInfo(object_type='dcim.region', field_name='termination', is_generic=True),
        'termination_site': RefInfo(object_type='dcim.site', field_name='termination', is_generic=True),
        'termination_site_group': RefInfo(object_type='dcim.sitegroup', field_name='termination', is_generic=True),
    },
    'circuits.circuittype': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'circuits.provider': {
        'accounts': RefInfo(object_type='circuits.provideraccount', field_name='accounts', is_many=True),
        'asns': RefInfo(object_type='ipam.asn', field_name='asns', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'circuits.provideraccount': {
        'provider': RefInfo(object_type='circuits.provider', field_name='provider'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'circuits.providernetwork': {
        'provider': RefInfo(object_type='circuits.provider', field_name='provider'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'circuits.virtualcircuit': {
        'provider_account': RefInfo(object_type='circuits.provideraccount', field_name='provider_account'),
        'provider_network': RefInfo(object_type='circuits.providernetwork', field_name='provider_network'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'type': RefInfo(object_type='circuits.virtualcircuittype', field_name='type'),
    },
    'circuits.virtualcircuittermination': {
        'interface': RefInfo(object_type='dcim.interface', field_name='interface'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'virtual_circuit': RefInfo(object_type='circuits.virtualcircuit', field_name='virtual_circuit'),
    },
    'circuits.virtualcircuittype': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.cable': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'dcim.cabletermination': {
        'cable': RefInfo(object_type='dcim.cable', field_name='cable'),
        'termination_circuit_termination': RefInfo(object_type='circuits.circuittermination', field_name='termination', is_generic=True),
        'termination_console_port': RefInfo(object_type='dcim.consoleport', field_name='termination', is_generic=True),
        'termination_console_server_port': RefInfo(object_type='dcim.consoleserverport', field_name='termination', is_generic=True),
        'termination_front_port': RefInfo(object_type='dcim.frontport', field_name='termination', is_generic=True),
        'termination_interface': RefInfo(object_type='dcim.interface', field_name='termination', is_generic=True),
        'termination_power_feed': RefInfo(object_type='dcim.powerfeed', field_name='termination', is_generic=True),
        'termination_power_outlet': RefInfo(object_type='dcim.poweroutlet', field_name='termination', is_generic=True),
        'termination_power_port': RefInfo(object_type='dcim.powerport', field_name='termination', is_generic=True),
        'termination_rear_port': RefInfo(object_type='dcim.rearport', field_name='termination', is_generic=True),
    },
    'dcim.consoleport': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.consoleserverport': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.device': {
        'cluster': RefInfo(object_type='virtualization.cluster', field_name='cluster'),
        'device_type': RefInfo(object_type='dcim.devicetype', field_name='device_type'),
        'location': RefInfo(object_type='dcim.location', field_name='location'),
        'oob_ip': RefInfo(object_type='ipam.ipaddress', field_name='oob_ip'),
        'platform': RefInfo(object_type='dcim.platform', field_name='platform'),
        'primary_ip4': RefInfo(object_type='ipam.ipaddress', field_name='primary_ip4'),
        'primary_ip6': RefInfo(object_type='ipam.ipaddress', field_name='primary_ip6'),
        'rack': RefInfo(object_type='dcim.rack', field_name='rack'),
        'role': RefInfo(object_type='dcim.devicerole', field_name='role'),
        'site': RefInfo(object_type='dcim.site', field_name='site'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'virtual_chassis': RefInfo(object_type='dcim.virtualchassis', field_name='virtual_chassis'),
    },
    'dcim.devicebay': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'installed_device': RefInfo(object_type='dcim.device', field_name='installed_device'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.devicerole': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.devicetype': {
        'default_platform': RefInfo(object_type='dcim.platform', field_name='default_platform'),
        'manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='manufacturer'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.frontport': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'rear_port': RefInfo(object_type='dcim.rearport', field_name='rear_port'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.interface': {
        'bridge': RefInfo(object_type='dcim.interface', field_name='bridge'),
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'lag': RefInfo(object_type='dcim.interface', field_name='lag'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'parent': RefInfo(object_type='dcim.interface', field_name='parent'),
        'primary_mac_address': RefInfo(object_type='dcim.macaddress', field_name='primary_mac_address'),
        'qinq_svlan': RefInfo(object_type='ipam.vlan', field_name='qinq_svlan'),
        'tagged_vlans': RefInfo(object_type='ipam.vlan', field_name='tagged_vlans', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'untagged_vlan': RefInfo(object_type='ipam.vlan', field_name='untagged_vlan'),
        'vdcs': RefInfo(object_type='dcim.virtualdevicecontext', field_name='vdcs', is_many=True),
        'vlan_translation_policy': RefInfo(object_type='ipam.vlantranslationpolicy', field_name='vlan_translation_policy'),
        'vrf': RefInfo(object_type='ipam.vrf', field_name='vrf'),
        'wireless_lans': RefInfo(object_type='wireless.wirelesslan', field_name='wireless_lans', is_many=True),
    },
    'dcim.inventoryitem': {
        'component_console_port': RefInfo(object_type='dcim.consoleport', field_name='component', is_generic=True),
        'component_console_server_port': RefInfo(object_type='dcim.consoleserverport', field_name='component', is_generic=True),
        'component_front_port': RefInfo(object_type='dcim.frontport', field_name='component', is_generic=True),
        'component_interface': RefInfo(object_type='dcim.interface', field_name='component', is_generic=True),
        'component_power_outlet': RefInfo(object_type='dcim.poweroutlet', field_name='component', is_generic=True),
        'component_power_port': RefInfo(object_type='dcim.powerport', field_name='component', is_generic=True),
        'component_rear_port': RefInfo(object_type='dcim.rearport', field_name='component', is_generic=True),
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='manufacturer'),
        'parent': RefInfo(object_type='dcim.inventoryitem', field_name='parent'),
        'role': RefInfo(object_type='dcim.inventoryitemrole', field_name='role'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.inventoryitemrole': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.location': {
        'parent': RefInfo(object_type='dcim.location', field_name='parent'),
        'site': RefInfo(object_type='dcim.site', field_name='site'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'dcim.macaddress': {
        'assigned_object_interface': RefInfo(object_type='dcim.interface', field_name='assigned_object', is_generic=True),
        'assigned_object_vm_interface': RefInfo(object_type='virtualization.vminterface', field_name='assigned_object', is_generic=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.manufacturer': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.module': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'module_bay': RefInfo(object_type='dcim.modulebay', field_name='module_bay'),
        'module_type': RefInfo(object_type='dcim.moduletype', field_name='module_type'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.modulebay': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'installed_module': RefInfo(object_type='dcim.module', field_name='installed_module'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.moduletype': {
        'manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='manufacturer'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.platform': {
        'manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='manufacturer'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.powerfeed': {
        'power_panel': RefInfo(object_type='dcim.powerpanel', field_name='power_panel'),
        'rack': RefInfo(object_type='dcim.rack', field_name='rack'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'dcim.poweroutlet': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'power_port': RefInfo(object_type='dcim.powerport', field_name='power_port'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.powerpanel': {
        'location': RefInfo(object_type='dcim.location', field_name='location'),
        'site': RefInfo(object_type='dcim.site', field_name='site'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.powerport': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.rack': {
        'location': RefInfo(object_type='dcim.location', field_name='location'),
        'rack_type': RefInfo(object_type='dcim.racktype', field_name='rack_type'),
        'role': RefInfo(object_type='dcim.rackrole', field_name='role'),
        'site': RefInfo(object_type='dcim.site', field_name='site'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'dcim.rackreservation': {
        'rack': RefInfo(object_type='dcim.rack', field_name='rack'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'dcim.rackrole': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.racktype': {
        'manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='manufacturer'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.rearport': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'module': RefInfo(object_type='dcim.module', field_name='module'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.region': {
        'parent': RefInfo(object_type='dcim.region', field_name='parent'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.site': {
        'asns': RefInfo(object_type='ipam.asn', field_name='asns', is_many=True),
        'group': RefInfo(object_type='dcim.sitegroup', field_name='group'),
        'region': RefInfo(object_type='dcim.region', field_name='region'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'dcim.sitegroup': {
        'parent': RefInfo(object_type='dcim.sitegroup', field_name='parent'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.virtualchassis': {
        'master': RefInfo(object_type='dcim.device', field_name='master'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'dcim.virtualdevicecontext': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'primary_ip4': RefInfo(object_type='ipam.ipaddress', field_name='primary_ip4'),
        'primary_ip6': RefInfo(object_type='ipam.ipaddress', field_name='primary_ip6'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'ipam.aggregate': {
        'rir': RefInfo(object_type='ipam.rir', field_name='rir'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'ipam.asn': {
        'rir': RefInfo(object_type='ipam.rir', field_name='rir'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'ipam.asnrange': {
        'rir': RefInfo(object_type='ipam.rir', field_name='rir'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'ipam.fhrpgroup': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'ipam.fhrpgroupassignment': {
        'group': RefInfo(object_type='ipam.fhrpgroup', field_name='group'),
        'interface_asn': RefInfo(object_type='ipam.asn', field_name='interface', is_generic=True),
        'interface_asn_range': RefInfo(object_type='ipam.asnrange', field_name='interface', is_generic=True),
        'interface_aggregate': RefInfo(object_type='ipam.aggregate', field_name='interface', is_generic=True),
        'interface_cable': RefInfo(object_type='dcim.cable', field_name='interface', is_generic=True),
        'interface_cable_path': RefInfo(object_type='dcim.cablepath', field_name='interface', is_generic=True),
        'interface_cable_termination': RefInfo(object_type='dcim.cabletermination', field_name='interface', is_generic=True),
        'interface_circuit': RefInfo(object_type='circuits.circuit', field_name='interface', is_generic=True),
        'interface_circuit_group': RefInfo(object_type='circuits.circuitgroup', field_name='interface', is_generic=True),
        'interface_circuit_group_assignment': RefInfo(object_type='circuits.circuitgroupassignment', field_name='interface', is_generic=True),
        'interface_circuit_termination': RefInfo(object_type='circuits.circuittermination', field_name='interface', is_generic=True),
        'interface_circuit_type': RefInfo(object_type='circuits.circuittype', field_name='interface', is_generic=True),
        'interface_cluster': RefInfo(object_type='virtualization.cluster', field_name='interface', is_generic=True),
        'interface_cluster_group': RefInfo(object_type='virtualization.clustergroup', field_name='interface', is_generic=True),
        'interface_cluster_type': RefInfo(object_type='virtualization.clustertype', field_name='interface', is_generic=True),
        'interface_console_port': RefInfo(object_type='dcim.consoleport', field_name='interface', is_generic=True),
        'interface_console_server_port': RefInfo(object_type='dcim.consoleserverport', field_name='interface', is_generic=True),
        'interface_contact': RefInfo(object_type='tenancy.contact', field_name='interface', is_generic=True),
        'interface_contact_assignment': RefInfo(object_type='tenancy.contactassignment', field_name='interface', is_generic=True),
        'interface_contact_group': RefInfo(object_type='tenancy.contactgroup', field_name='interface', is_generic=True),
        'interface_contact_role': RefInfo(object_type='tenancy.contactrole', field_name='interface', is_generic=True),
        'interface_device': RefInfo(object_type='dcim.device', field_name='interface', is_generic=True),
        'interface_device_bay': RefInfo(object_type='dcim.devicebay', field_name='interface', is_generic=True),
        'interface_device_role': RefInfo(object_type='dcim.devicerole', field_name='interface', is_generic=True),
        'interface_device_type': RefInfo(object_type='dcim.devicetype', field_name='interface', is_generic=True),
        'interface_fhrp_group': RefInfo(object_type='ipam.fhrpgroup', field_name='interface', is_generic=True),
        'interface_fhrp_group_assignment': RefInfo(object_type='ipam.fhrpgroupassignment', field_name='interface', is_generic=True),
        'interface_front_port': RefInfo(object_type='dcim.frontport', field_name='interface', is_generic=True),
        'interface_ike_policy': RefInfo(object_type='vpn.ikepolicy', field_name='interface', is_generic=True),
        'interface_ike_proposal': RefInfo(object_type='vpn.ikeproposal', field_name='interface', is_generic=True),
        'interface_ip_address': RefInfo(object_type='ipam.ipaddress', field_name='interface', is_generic=True),
        'interface_ip_range': RefInfo(object_type='ipam.iprange', field_name='interface', is_generic=True),
        'interface_ip_sec_policy': RefInfo(object_type='vpn.ipsecpolicy', field_name='interface', is_generic=True),
        'interface_ip_sec_profile': RefInfo(object_type='vpn.ipsecprofile', field_name='interface', is_generic=True),
        'interface_ip_sec_proposal': RefInfo(object_type='vpn.ipsecproposal', field_name='interface', is_generic=True),
        'interface_interface': RefInfo(object_type='dcim.interface', field_name='interface', is_generic=True),
        'interface_inventory_item': RefInfo(object_type='dcim.inventoryitem', field_name='interface', is_generic=True),
        'interface_inventory_item_role': RefInfo(object_type='dcim.inventoryitemrole', field_name='interface', is_generic=True),
        'interface_l2vpn': RefInfo(object_type='vpn.l2vpn', field_name='interface', is_generic=True),
        'interface_l2vpn_termination': RefInfo(object_type='vpn.l2vpntermination', field_name='interface', is_generic=True),
        'interface_location': RefInfo(object_type='dcim.location', field_name='interface', is_generic=True),
        'interface_mac_address': RefInfo(object_type='dcim.macaddress', field_name='interface', is_generic=True),
        'interface_manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='interface', is_generic=True),
        'interface_module': RefInfo(object_type='dcim.module', field_name='interface', is_generic=True),
        'interface_module_bay': RefInfo(object_type='dcim.modulebay', field_name='interface', is_generic=True),
        'interface_module_type': RefInfo(object_type='dcim.moduletype', field_name='interface', is_generic=True),
        'interface_platform': RefInfo(object_type='dcim.platform', field_name='interface', is_generic=True),
        'interface_power_feed': RefInfo(object_type='dcim.powerfeed', field_name='interface', is_generic=True),
        'interface_power_outlet': RefInfo(object_type='dcim.poweroutlet', field_name='interface', is_generic=True),
        'interface_power_panel': RefInfo(object_type='dcim.powerpanel', field_name='interface', is_generic=True),
        'interface_power_port': RefInfo(object_type='dcim.powerport', field_name='interface', is_generic=True),
        'interface_prefix': RefInfo(object_type='ipam.prefix', field_name='interface', is_generic=True),
        'interface_provider': RefInfo(object_type='circuits.provider', field_name='interface', is_generic=True),
        'interface_provider_account': RefInfo(object_type='circuits.provideraccount', field_name='interface', is_generic=True),
        'interface_provider_network': RefInfo(object_type='circuits.providernetwork', field_name='interface', is_generic=True),
        'interface_rir': RefInfo(object_type='ipam.rir', field_name='interface', is_generic=True),
        'interface_rack': RefInfo(object_type='dcim.rack', field_name='interface', is_generic=True),
        'interface_rack_reservation': RefInfo(object_type='dcim.rackreservation', field_name='interface', is_generic=True),
        'interface_rack_role': RefInfo(object_type='dcim.rackrole', field_name='interface', is_generic=True),
        'interface_rack_type': RefInfo(object_type='dcim.racktype', field_name='interface', is_generic=True),
        'interface_rear_port': RefInfo(object_type='dcim.rearport', field_name='interface', is_generic=True),
        'interface_region': RefInfo(object_type='dcim.region', field_name='interface', is_generic=True),
        'interface_role': RefInfo(object_type='ipam.role', field_name='interface', is_generic=True),
        'interface_route_target': RefInfo(object_type='ipam.routetarget', field_name='interface', is_generic=True),
        'interface_service': RefInfo(object_type='ipam.service', field_name='interface', is_generic=True),
        'interface_site': RefInfo(object_type='dcim.site', field_name='interface', is_generic=True),
        'interface_site_group': RefInfo(object_type='dcim.sitegroup', field_name='interface', is_generic=True),
        'interface_tag': RefInfo(object_type='extras.tag', field_name='interface', is_generic=True),
        'interface_tenant': RefInfo(object_type='tenancy.tenant', field_name='interface', is_generic=True),
        'interface_tenant_group': RefInfo(object_type='tenancy.tenantgroup', field_name='interface', is_generic=True),
        'interface_tunnel': RefInfo(object_type='vpn.tunnel', field_name='interface', is_generic=True),
        'interface_tunnel_group': RefInfo(object_type='vpn.tunnelgroup', field_name='interface', is_generic=True),
        'interface_tunnel_termination': RefInfo(object_type='vpn.tunneltermination', field_name='interface', is_generic=True),
        'interface_vlan': RefInfo(object_type='ipam.vlan', field_name='interface', is_generic=True),
        'interface_vlan_group': RefInfo(object_type='ipam.vlangroup', field_name='interface', is_generic=True),
        'interface_vlan_translation_policy': RefInfo(object_type='ipam.vlantranslationpolicy', field_name='interface', is_generic=True),
        'interface_vlan_translation_rule': RefInfo(object_type='ipam.vlantranslationrule', field_name='interface', is_generic=True),
        'interface_vm_interface': RefInfo(object_type='virtualization.vminterface', field_name='interface', is_generic=True),
        'interface_vrf': RefInfo(object_type='ipam.vrf', field_name='interface', is_generic=True),
        'interface_virtual_chassis': RefInfo(object_type='dcim.virtualchassis', field_name='interface', is_generic=True),
        'interface_virtual_circuit': RefInfo(object_type='circuits.virtualcircuit', field_name='interface', is_generic=True),
        'interface_virtual_circuit_termination': RefInfo(object_type='circuits.virtualcircuittermination', field_name='interface', is_generic=True),
        'interface_virtual_circuit_type': RefInfo(object_type='circuits.virtualcircuittype', field_name='interface', is_generic=True),
        'interface_virtual_device_context': RefInfo(object_type='dcim.virtualdevicecontext', field_name='interface', is_generic=True),
        'interface_virtual_disk': RefInfo(object_type='virtualization.virtualdisk', field_name='interface', is_generic=True),
        'interface_virtual_machine': RefInfo(object_type='virtualization.virtualmachine', field_name='interface', is_generic=True),
        'interface_wireless_lan': RefInfo(object_type='wireless.wirelesslan', field_name='interface', is_generic=True),
        'interface_wireless_lan_group': RefInfo(object_type='wireless.wirelesslangroup', field_name='interface', is_generic=True),
        'interface_wireless_link': RefInfo(object_type='wireless.wirelesslink', field_name='interface', is_generic=True),
    },
    'ipam.ipaddress': {
        'assigned_object_fhrp_group': RefInfo(object_type='ipam.fhrpgroup', field_name='assigned_object', is_generic=True),
        'assigned_object_interface': RefInfo(object_type='dcim.interface', field_name='assigned_object', is_generic=True),
        'assigned_object_vm_interface': RefInfo(object_type='virtualization.vminterface', field_name='assigned_object', is_generic=True),
        'nat_inside': RefInfo(object_type='ipam.ipaddress', field_name='nat_inside'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'vrf': RefInfo(object_type='ipam.vrf', field_name='vrf'),
    },
    'ipam.iprange': {
        'role': RefInfo(object_type='ipam.role', field_name='role'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'vrf': RefInfo(object_type='ipam.vrf', field_name='vrf'),
    },
    'ipam.prefix': {
        'role': RefInfo(object_type='ipam.role', field_name='role'),
        'scope_location': RefInfo(object_type='dcim.location', field_name='scope', is_generic=True),
        'scope_region': RefInfo(object_type='dcim.region', field_name='scope', is_generic=True),
        'scope_site': RefInfo(object_type='dcim.site', field_name='scope', is_generic=True),
        'scope_site_group': RefInfo(object_type='dcim.sitegroup', field_name='scope', is_generic=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'vlan': RefInfo(object_type='ipam.vlan', field_name='vlan'),
        'vrf': RefInfo(object_type='ipam.vrf', field_name='vrf'),
    },
    'ipam.rir': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'ipam.role': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'ipam.routetarget': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'ipam.service': {
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'ipaddresses': RefInfo(object_type='ipam.ipaddress', field_name='ipaddresses', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'virtual_machine': RefInfo(object_type='virtualization.virtualmachine', field_name='virtual_machine'),
    },
    'ipam.vlan': {
        'group': RefInfo(object_type='ipam.vlangroup', field_name='group'),
        'qinq_svlan': RefInfo(object_type='ipam.vlan', field_name='qinq_svlan'),
        'role': RefInfo(object_type='ipam.role', field_name='role'),
        'site': RefInfo(object_type='dcim.site', field_name='site'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'ipam.vlangroup': {
        'scope_cluster': RefInfo(object_type='virtualization.cluster', field_name='scope', is_generic=True),
        'scope_cluster_group': RefInfo(object_type='virtualization.clustergroup', field_name='scope', is_generic=True),
        'scope_location': RefInfo(object_type='dcim.location', field_name='scope', is_generic=True),
        'scope_rack': RefInfo(object_type='dcim.rack', field_name='scope', is_generic=True),
        'scope_region': RefInfo(object_type='dcim.region', field_name='scope', is_generic=True),
        'scope_site': RefInfo(object_type='dcim.site', field_name='scope', is_generic=True),
        'scope_site_group': RefInfo(object_type='dcim.sitegroup', field_name='scope', is_generic=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'ipam.vlantranslationrule': {
        'policy': RefInfo(object_type='ipam.vlantranslationpolicy', field_name='policy'),
    },
    'ipam.vrf': {
        'export_targets': RefInfo(object_type='ipam.routetarget', field_name='export_targets', is_many=True),
        'import_targets': RefInfo(object_type='ipam.routetarget', field_name='import_targets', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'tenancy.contact': {
        'group': RefInfo(object_type='tenancy.contactgroup', field_name='group'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'tenancy.contactassignment': {
        'contact': RefInfo(object_type='tenancy.contact', field_name='contact'),
        'object_asn': RefInfo(object_type='ipam.asn', field_name='object', is_generic=True),
        'object_asn_range': RefInfo(object_type='ipam.asnrange', field_name='object', is_generic=True),
        'object_aggregate': RefInfo(object_type='ipam.aggregate', field_name='object', is_generic=True),
        'object_cable': RefInfo(object_type='dcim.cable', field_name='object', is_generic=True),
        'object_cable_path': RefInfo(object_type='dcim.cablepath', field_name='object', is_generic=True),
        'object_cable_termination': RefInfo(object_type='dcim.cabletermination', field_name='object', is_generic=True),
        'object_circuit': RefInfo(object_type='circuits.circuit', field_name='object', is_generic=True),
        'object_circuit_group': RefInfo(object_type='circuits.circuitgroup', field_name='object', is_generic=True),
        'object_circuit_group_assignment': RefInfo(object_type='circuits.circuitgroupassignment', field_name='object', is_generic=True),
        'object_circuit_termination': RefInfo(object_type='circuits.circuittermination', field_name='object', is_generic=True),
        'object_circuit_type': RefInfo(object_type='circuits.circuittype', field_name='object', is_generic=True),
        'object_cluster': RefInfo(object_type='virtualization.cluster', field_name='object', is_generic=True),
        'object_cluster_group': RefInfo(object_type='virtualization.clustergroup', field_name='object', is_generic=True),
        'object_cluster_type': RefInfo(object_type='virtualization.clustertype', field_name='object', is_generic=True),
        'object_console_port': RefInfo(object_type='dcim.consoleport', field_name='object', is_generic=True),
        'object_console_server_port': RefInfo(object_type='dcim.consoleserverport', field_name='object', is_generic=True),
        'object_contact': RefInfo(object_type='tenancy.contact', field_name='object', is_generic=True),
        'object_contact_assignment': RefInfo(object_type='tenancy.contactassignment', field_name='object', is_generic=True),
        'object_contact_group': RefInfo(object_type='tenancy.contactgroup', field_name='object', is_generic=True),
        'object_contact_role': RefInfo(object_type='tenancy.contactrole', field_name='object', is_generic=True),
        'object_device': RefInfo(object_type='dcim.device', field_name='object', is_generic=True),
        'object_device_bay': RefInfo(object_type='dcim.devicebay', field_name='object', is_generic=True),
        'object_device_role': RefInfo(object_type='dcim.devicerole', field_name='object', is_generic=True),
        'object_device_type': RefInfo(object_type='dcim.devicetype', field_name='object', is_generic=True),
        'object_fhrp_group': RefInfo(object_type='ipam.fhrpgroup', field_name='object', is_generic=True),
        'object_fhrp_group_assignment': RefInfo(object_type='ipam.fhrpgroupassignment', field_name='object', is_generic=True),
        'object_front_port': RefInfo(object_type='dcim.frontport', field_name='object', is_generic=True),
        'object_ike_policy': RefInfo(object_type='vpn.ikepolicy', field_name='object', is_generic=True),
        'object_ike_proposal': RefInfo(object_type='vpn.ikeproposal', field_name='object', is_generic=True),
        'object_ip_address': RefInfo(object_type='ipam.ipaddress', field_name='object', is_generic=True),
        'object_ip_range': RefInfo(object_type='ipam.iprange', field_name='object', is_generic=True),
        'object_ip_sec_policy': RefInfo(object_type='vpn.ipsecpolicy', field_name='object', is_generic=True),
        'object_ip_sec_profile': RefInfo(object_type='vpn.ipsecprofile', field_name='object', is_generic=True),
        'object_ip_sec_proposal': RefInfo(object_type='vpn.ipsecproposal', field_name='object', is_generic=True),
        'object_interface': RefInfo(object_type='dcim.interface', field_name='object', is_generic=True),
        'object_inventory_item': RefInfo(object_type='dcim.inventoryitem', field_name='object', is_generic=True),
        'object_inventory_item_role': RefInfo(object_type='dcim.inventoryitemrole', field_name='object', is_generic=True),
        'object_l2vpn': RefInfo(object_type='vpn.l2vpn', field_name='object', is_generic=True),
        'object_l2vpn_termination': RefInfo(object_type='vpn.l2vpntermination', field_name='object', is_generic=True),
        'object_location': RefInfo(object_type='dcim.location', field_name='object', is_generic=True),
        'object_mac_address': RefInfo(object_type='dcim.macaddress', field_name='object', is_generic=True),
        'object_manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='object', is_generic=True),
        'object_module': RefInfo(object_type='dcim.module', field_name='object', is_generic=True),
        'object_module_bay': RefInfo(object_type='dcim.modulebay', field_name='object', is_generic=True),
        'object_module_type': RefInfo(object_type='dcim.moduletype', field_name='object', is_generic=True),
        'object_platform': RefInfo(object_type='dcim.platform', field_name='object', is_generic=True),
        'object_power_feed': RefInfo(object_type='dcim.powerfeed', field_name='object', is_generic=True),
        'object_power_outlet': RefInfo(object_type='dcim.poweroutlet', field_name='object', is_generic=True),
        'object_power_panel': RefInfo(object_type='dcim.powerpanel', field_name='object', is_generic=True),
        'object_power_port': RefInfo(object_type='dcim.powerport', field_name='object', is_generic=True),
        'object_prefix': RefInfo(object_type='ipam.prefix', field_name='object', is_generic=True),
        'object_provider': RefInfo(object_type='circuits.provider', field_name='object', is_generic=True),
        'object_provider_account': RefInfo(object_type='circuits.provideraccount', field_name='object', is_generic=True),
        'object_provider_network': RefInfo(object_type='circuits.providernetwork', field_name='object', is_generic=True),
        'object_rir': RefInfo(object_type='ipam.rir', field_name='object', is_generic=True),
        'object_rack': RefInfo(object_type='dcim.rack', field_name='object', is_generic=True),
        'object_rack_reservation': RefInfo(object_type='dcim.rackreservation', field_name='object', is_generic=True),
        'object_rack_role': RefInfo(object_type='dcim.rackrole', field_name='object', is_generic=True),
        'object_rack_type': RefInfo(object_type='dcim.racktype', field_name='object', is_generic=True),
        'object_rear_port': RefInfo(object_type='dcim.rearport', field_name='object', is_generic=True),
        'object_region': RefInfo(object_type='dcim.region', field_name='object', is_generic=True),
        'object_role': RefInfo(object_type='ipam.role', field_name='object', is_generic=True),
        'object_route_target': RefInfo(object_type='ipam.routetarget', field_name='object', is_generic=True),
        'object_service': RefInfo(object_type='ipam.service', field_name='object', is_generic=True),
        'object_site': RefInfo(object_type='dcim.site', field_name='object', is_generic=True),
        'object_site_group': RefInfo(object_type='dcim.sitegroup', field_name='object', is_generic=True),
        'object_tag': RefInfo(object_type='extras.tag', field_name='object', is_generic=True),
        'object_tenant': RefInfo(object_type='tenancy.tenant', field_name='object', is_generic=True),
        'object_tenant_group': RefInfo(object_type='tenancy.tenantgroup', field_name='object', is_generic=True),
        'object_tunnel': RefInfo(object_type='vpn.tunnel', field_name='object', is_generic=True),
        'object_tunnel_group': RefInfo(object_type='vpn.tunnelgroup', field_name='object', is_generic=True),
        'object_tunnel_termination': RefInfo(object_type='vpn.tunneltermination', field_name='object', is_generic=True),
        'object_vlan': RefInfo(object_type='ipam.vlan', field_name='object', is_generic=True),
        'object_vlan_group': RefInfo(object_type='ipam.vlangroup', field_name='object', is_generic=True),
        'object_vlan_translation_policy': RefInfo(object_type='ipam.vlantranslationpolicy', field_name='object', is_generic=True),
        'object_vlan_translation_rule': RefInfo(object_type='ipam.vlantranslationrule', field_name='object', is_generic=True),
        'object_vm_interface': RefInfo(object_type='virtualization.vminterface', field_name='object', is_generic=True),
        'object_vrf': RefInfo(object_type='ipam.vrf', field_name='object', is_generic=True),
        'object_virtual_chassis': RefInfo(object_type='dcim.virtualchassis', field_name='object', is_generic=True),
        'object_virtual_circuit': RefInfo(object_type='circuits.virtualcircuit', field_name='object', is_generic=True),
        'object_virtual_circuit_termination': RefInfo(object_type='circuits.virtualcircuittermination', field_name='object', is_generic=True),
        'object_virtual_circuit_type': RefInfo(object_type='circuits.virtualcircuittype', field_name='object', is_generic=True),
        'object_virtual_device_context': RefInfo(object_type='dcim.virtualdevicecontext', field_name='object', is_generic=True),
        'object_virtual_disk': RefInfo(object_type='virtualization.virtualdisk', field_name='object', is_generic=True),
        'object_virtual_machine': RefInfo(object_type='virtualization.virtualmachine', field_name='object', is_generic=True),
        'object_wireless_lan': RefInfo(object_type='wireless.wirelesslan', field_name='object', is_generic=True),
        'object_wireless_lan_group': RefInfo(object_type='wireless.wirelesslangroup', field_name='object', is_generic=True),
        'object_wireless_link': RefInfo(object_type='wireless.wirelesslink', field_name='object', is_generic=True),
        'role': RefInfo(object_type='tenancy.contactrole', field_name='role'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'tenancy.contactgroup': {
        'parent': RefInfo(object_type='tenancy.contactgroup', field_name='parent'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'tenancy.contactrole': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'tenancy.tenant': {
        'group': RefInfo(object_type='tenancy.tenantgroup', field_name='group'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'tenancy.tenantgroup': {
        'parent': RefInfo(object_type='tenancy.tenantgroup', field_name='parent'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'virtualization.cluster': {
        'group': RefInfo(object_type='virtualization.clustergroup', field_name='group'),
        'scope_location': RefInfo(object_type='dcim.location', field_name='scope', is_generic=True),
        'scope_region': RefInfo(object_type='dcim.region', field_name='scope', is_generic=True),
        'scope_site': RefInfo(object_type='dcim.site', field_name='scope', is_generic=True),
        'scope_site_group': RefInfo(object_type='dcim.sitegroup', field_name='scope', is_generic=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'type': RefInfo(object_type='virtualization.clustertype', field_name='type'),
    },
    'virtualization.clustergroup': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'virtualization.clustertype': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'virtualization.virtualdisk': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'virtual_machine': RefInfo(object_type='virtualization.virtualmachine', field_name='virtual_machine'),
    },
    'virtualization.virtualmachine': {
        'cluster': RefInfo(object_type='virtualization.cluster', field_name='cluster'),
        'device': RefInfo(object_type='dcim.device', field_name='device'),
        'platform': RefInfo(object_type='dcim.platform', field_name='platform'),
        'primary_ip4': RefInfo(object_type='ipam.ipaddress', field_name='primary_ip4'),
        'primary_ip6': RefInfo(object_type='ipam.ipaddress', field_name='primary_ip6'),
        'role': RefInfo(object_type='dcim.devicerole', field_name='role'),
        'site': RefInfo(object_type='dcim.site', field_name='site'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'virtualization.vminterface': {
        'bridge': RefInfo(object_type='virtualization.vminterface', field_name='bridge'),
        'parent': RefInfo(object_type='virtualization.vminterface', field_name='parent'),
        'primary_mac_address': RefInfo(object_type='dcim.macaddress', field_name='primary_mac_address'),
        'qinq_svlan': RefInfo(object_type='ipam.vlan', field_name='qinq_svlan'),
        'tagged_vlans': RefInfo(object_type='ipam.vlan', field_name='tagged_vlans', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'untagged_vlan': RefInfo(object_type='ipam.vlan', field_name='untagged_vlan'),
        'virtual_machine': RefInfo(object_type='virtualization.virtualmachine', field_name='virtual_machine'),
        'vlan_translation_policy': RefInfo(object_type='ipam.vlantranslationpolicy', field_name='vlan_translation_policy'),
        'vrf': RefInfo(object_type='ipam.vrf', field_name='vrf'),
    },
    'vpn.ikepolicy': {
        'proposals': RefInfo(object_type='vpn.ikeproposal', field_name='proposals', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'vpn.ikeproposal': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'vpn.ipsecpolicy': {
        'proposals': RefInfo(object_type='vpn.ipsecproposal', field_name='proposals', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'vpn.ipsecprofile': {
        'ike_policy': RefInfo(object_type='vpn.ikepolicy', field_name='ike_policy'),
        'ipsec_policy': RefInfo(object_type='vpn.ipsecpolicy', field_name='ipsec_policy'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'vpn.ipsecproposal': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'vpn.l2vpn': {
        'export_targets': RefInfo(object_type='ipam.routetarget', field_name='export_targets', is_many=True),
        'import_targets': RefInfo(object_type='ipam.routetarget', field_name='import_targets', is_many=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'vpn.l2vpntermination': {
        'assigned_object_interface': RefInfo(object_type='dcim.interface', field_name='assigned_object', is_generic=True),
        'assigned_object_vlan': RefInfo(object_type='ipam.vlan', field_name='assigned_object', is_generic=True),
        'assigned_object_vm_interface': RefInfo(object_type='virtualization.vminterface', field_name='assigned_object', is_generic=True),
        'l2vpn': RefInfo(object_type='vpn.l2vpn', field_name='l2vpn'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'vpn.tunnel': {
        'group': RefInfo(object_type='vpn.tunnelgroup', field_name='group'),
        'ipsec_profile': RefInfo(object_type='vpn.ipsecprofile', field_name='ipsec_profile'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
    'vpn.tunnelgroup': {
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'vpn.tunneltermination': {
        'outside_ip': RefInfo(object_type='ipam.ipaddress', field_name='outside_ip'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'termination_asn': RefInfo(object_type='ipam.asn', field_name='termination', is_generic=True),
        'termination_asn_range': RefInfo(object_type='ipam.asnrange', field_name='termination', is_generic=True),
        'termination_aggregate': RefInfo(object_type='ipam.aggregate', field_name='termination', is_generic=True),
        'termination_cable': RefInfo(object_type='dcim.cable', field_name='termination', is_generic=True),
        'termination_cable_path': RefInfo(object_type='dcim.cablepath', field_name='termination', is_generic=True),
        'termination_cable_termination': RefInfo(object_type='dcim.cabletermination', field_name='termination', is_generic=True),
        'termination_circuit': RefInfo(object_type='circuits.circuit', field_name='termination', is_generic=True),
        'termination_circuit_group': RefInfo(object_type='circuits.circuitgroup', field_name='termination', is_generic=True),
        'termination_circuit_group_assignment': RefInfo(object_type='circuits.circuitgroupassignment', field_name='termination', is_generic=True),
        'termination_circuit_termination': RefInfo(object_type='circuits.circuittermination', field_name='termination', is_generic=True),
        'termination_circuit_type': RefInfo(object_type='circuits.circuittype', field_name='termination', is_generic=True),
        'termination_cluster': RefInfo(object_type='virtualization.cluster', field_name='termination', is_generic=True),
        'termination_cluster_group': RefInfo(object_type='virtualization.clustergroup', field_name='termination', is_generic=True),
        'termination_cluster_type': RefInfo(object_type='virtualization.clustertype', field_name='termination', is_generic=True),
        'termination_console_port': RefInfo(object_type='dcim.consoleport', field_name='termination', is_generic=True),
        'termination_console_server_port': RefInfo(object_type='dcim.consoleserverport', field_name='termination', is_generic=True),
        'termination_contact': RefInfo(object_type='tenancy.contact', field_name='termination', is_generic=True),
        'termination_contact_assignment': RefInfo(object_type='tenancy.contactassignment', field_name='termination', is_generic=True),
        'termination_contact_group': RefInfo(object_type='tenancy.contactgroup', field_name='termination', is_generic=True),
        'termination_contact_role': RefInfo(object_type='tenancy.contactrole', field_name='termination', is_generic=True),
        'termination_device': RefInfo(object_type='dcim.device', field_name='termination', is_generic=True),
        'termination_device_bay': RefInfo(object_type='dcim.devicebay', field_name='termination', is_generic=True),
        'termination_device_role': RefInfo(object_type='dcim.devicerole', field_name='termination', is_generic=True),
        'termination_device_type': RefInfo(object_type='dcim.devicetype', field_name='termination', is_generic=True),
        'termination_fhrp_group': RefInfo(object_type='ipam.fhrpgroup', field_name='termination', is_generic=True),
        'termination_fhrp_group_assignment': RefInfo(object_type='ipam.fhrpgroupassignment', field_name='termination', is_generic=True),
        'termination_front_port': RefInfo(object_type='dcim.frontport', field_name='termination', is_generic=True),
        'termination_ike_policy': RefInfo(object_type='vpn.ikepolicy', field_name='termination', is_generic=True),
        'termination_ike_proposal': RefInfo(object_type='vpn.ikeproposal', field_name='termination', is_generic=True),
        'termination_ip_address': RefInfo(object_type='ipam.ipaddress', field_name='termination', is_generic=True),
        'termination_ip_range': RefInfo(object_type='ipam.iprange', field_name='termination', is_generic=True),
        'termination_ip_sec_policy': RefInfo(object_type='vpn.ipsecpolicy', field_name='termination', is_generic=True),
        'termination_ip_sec_profile': RefInfo(object_type='vpn.ipsecprofile', field_name='termination', is_generic=True),
        'termination_ip_sec_proposal': RefInfo(object_type='vpn.ipsecproposal', field_name='termination', is_generic=True),
        'termination_interface': RefInfo(object_type='dcim.interface', field_name='termination', is_generic=True),
        'termination_inventory_item': RefInfo(object_type='dcim.inventoryitem', field_name='termination', is_generic=True),
        'termination_inventory_item_role': RefInfo(object_type='dcim.inventoryitemrole', field_name='termination', is_generic=True),
        'termination_l2vpn': RefInfo(object_type='vpn.l2vpn', field_name='termination', is_generic=True),
        'termination_l2vpn_termination': RefInfo(object_type='vpn.l2vpntermination', field_name='termination', is_generic=True),
        'termination_location': RefInfo(object_type='dcim.location', field_name='termination', is_generic=True),
        'termination_mac_address': RefInfo(object_type='dcim.macaddress', field_name='termination', is_generic=True),
        'termination_manufacturer': RefInfo(object_type='dcim.manufacturer', field_name='termination', is_generic=True),
        'termination_module': RefInfo(object_type='dcim.module', field_name='termination', is_generic=True),
        'termination_module_bay': RefInfo(object_type='dcim.modulebay', field_name='termination', is_generic=True),
        'termination_module_type': RefInfo(object_type='dcim.moduletype', field_name='termination', is_generic=True),
        'termination_platform': RefInfo(object_type='dcim.platform', field_name='termination', is_generic=True),
        'termination_power_feed': RefInfo(object_type='dcim.powerfeed', field_name='termination', is_generic=True),
        'termination_power_outlet': RefInfo(object_type='dcim.poweroutlet', field_name='termination', is_generic=True),
        'termination_power_panel': RefInfo(object_type='dcim.powerpanel', field_name='termination', is_generic=True),
        'termination_power_port': RefInfo(object_type='dcim.powerport', field_name='termination', is_generic=True),
        'termination_prefix': RefInfo(object_type='ipam.prefix', field_name='termination', is_generic=True),
        'termination_provider': RefInfo(object_type='circuits.provider', field_name='termination', is_generic=True),
        'termination_provider_account': RefInfo(object_type='circuits.provideraccount', field_name='termination', is_generic=True),
        'termination_provider_network': RefInfo(object_type='circuits.providernetwork', field_name='termination', is_generic=True),
        'termination_rir': RefInfo(object_type='ipam.rir', field_name='termination', is_generic=True),
        'termination_rack': RefInfo(object_type='dcim.rack', field_name='termination', is_generic=True),
        'termination_rack_reservation': RefInfo(object_type='dcim.rackreservation', field_name='termination', is_generic=True),
        'termination_rack_role': RefInfo(object_type='dcim.rackrole', field_name='termination', is_generic=True),
        'termination_rack_type': RefInfo(object_type='dcim.racktype', field_name='termination', is_generic=True),
        'termination_rear_port': RefInfo(object_type='dcim.rearport', field_name='termination', is_generic=True),
        'termination_region': RefInfo(object_type='dcim.region', field_name='termination', is_generic=True),
        'termination_role': RefInfo(object_type='ipam.role', field_name='termination', is_generic=True),
        'termination_route_target': RefInfo(object_type='ipam.routetarget', field_name='termination', is_generic=True),
        'termination_service': RefInfo(object_type='ipam.service', field_name='termination', is_generic=True),
        'termination_site': RefInfo(object_type='dcim.site', field_name='termination', is_generic=True),
        'termination_site_group': RefInfo(object_type='dcim.sitegroup', field_name='termination', is_generic=True),
        'termination_tag': RefInfo(object_type='extras.tag', field_name='termination', is_generic=True),
        'termination_tenant': RefInfo(object_type='tenancy.tenant', field_name='termination', is_generic=True),
        'termination_tenant_group': RefInfo(object_type='tenancy.tenantgroup', field_name='termination', is_generic=True),
        'termination_tunnel': RefInfo(object_type='vpn.tunnel', field_name='termination', is_generic=True),
        'termination_tunnel_group': RefInfo(object_type='vpn.tunnelgroup', field_name='termination', is_generic=True),
        'termination_tunnel_termination': RefInfo(object_type='vpn.tunneltermination', field_name='termination', is_generic=True),
        'termination_vlan': RefInfo(object_type='ipam.vlan', field_name='termination', is_generic=True),
        'termination_vlan_group': RefInfo(object_type='ipam.vlangroup', field_name='termination', is_generic=True),
        'termination_vlan_translation_policy': RefInfo(object_type='ipam.vlantranslationpolicy', field_name='termination', is_generic=True),
        'termination_vlan_translation_rule': RefInfo(object_type='ipam.vlantranslationrule', field_name='termination', is_generic=True),
        'termination_vm_interface': RefInfo(object_type='virtualization.vminterface', field_name='termination', is_generic=True),
        'termination_vrf': RefInfo(object_type='ipam.vrf', field_name='termination', is_generic=True),
        'termination_virtual_chassis': RefInfo(object_type='dcim.virtualchassis', field_name='termination', is_generic=True),
        'termination_virtual_circuit': RefInfo(object_type='circuits.virtualcircuit', field_name='termination', is_generic=True),
        'termination_virtual_circuit_termination': RefInfo(object_type='circuits.virtualcircuittermination', field_name='termination', is_generic=True),
        'termination_virtual_circuit_type': RefInfo(object_type='circuits.virtualcircuittype', field_name='termination', is_generic=True),
        'termination_virtual_device_context': RefInfo(object_type='dcim.virtualdevicecontext', field_name='termination', is_generic=True),
        'termination_virtual_disk': RefInfo(object_type='virtualization.virtualdisk', field_name='termination', is_generic=True),
        'termination_virtual_machine': RefInfo(object_type='virtualization.virtualmachine', field_name='termination', is_generic=True),
        'termination_wireless_lan': RefInfo(object_type='wireless.wirelesslan', field_name='termination', is_generic=True),
        'termination_wireless_lan_group': RefInfo(object_type='wireless.wirelesslangroup', field_name='termination', is_generic=True),
        'termination_wireless_link': RefInfo(object_type='wireless.wirelesslink', field_name='termination', is_generic=True),
        'tunnel': RefInfo(object_type='vpn.tunnel', field_name='tunnel'),
    },
    'wireless.wirelesslan': {
        'group': RefInfo(object_type='wireless.wirelesslangroup', field_name='group'),
        'scope_location': RefInfo(object_type='dcim.location', field_name='scope', is_generic=True),
        'scope_region': RefInfo(object_type='dcim.region', field_name='scope', is_generic=True),
        'scope_site': RefInfo(object_type='dcim.site', field_name='scope', is_generic=True),
        'scope_site_group': RefInfo(object_type='dcim.sitegroup', field_name='scope', is_generic=True),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
        'vlan': RefInfo(object_type='ipam.vlan', field_name='vlan'),
    },
    'wireless.wirelesslangroup': {
        'parent': RefInfo(object_type='wireless.wirelesslangroup', field_name='parent'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
    },
    'wireless.wirelesslink': {
        'interface_a': RefInfo(object_type='dcim.interface', field_name='interface_a'),
        'interface_b': RefInfo(object_type='dcim.interface', field_name='interface_b'),
        'tags': RefInfo(object_type='extras.tag', field_name='tags', is_many=True),
        'tenant': RefInfo(object_type='tenancy.tenant', field_name='tenant'),
    },
}

def get_json_ref_info(object_type: str|Type[models.Model], json_field_name: str) -> RefInfo|None:
    if not isinstance(object_type, str):
        object_type = get_object_type(object_type)
    return _JSON_REF_INFO.get(object_type, {}).get(json_field_name)

_LEGAL_FIELDS = {
    'circuits.circuit': frozenset(['assignments', 'cid', 'comments', 'commit_rate', 'custom_fields', 'description', 'distance', 'distance_unit', 'install_date', 'provider', 'provider_account', 'status', 'tags', 'tenant', 'termination_date', 'type']),
    'circuits.circuitgroup': frozenset(['custom_fields', 'description', 'name', 'slug', 'tags', 'tenant']),
    'circuits.circuitgroupassignment': frozenset(['group', 'member_id', 'member_type', 'priority', 'tags']),
    'circuits.circuittermination': frozenset(['circuit', 'custom_fields', 'description', 'mark_connected', 'port_speed', 'pp_info', 'tags', 'term_side', 'termination_id', 'termination_type', 'upstream_speed', 'xconnect_id']),
    'circuits.circuittype': frozenset(['color', 'custom_fields', 'description', 'name', 'slug', 'tags']),
    'circuits.provider': frozenset(['accounts', 'asns', 'comments', 'custom_fields', 'description', 'name', 'slug', 'tags']),
    'circuits.provideraccount': frozenset(['account', 'comments', 'custom_fields', 'description', 'name', 'provider', 'tags']),
    'circuits.providernetwork': frozenset(['comments', 'custom_fields', 'description', 'name', 'provider', 'service_id', 'tags']),
    'circuits.virtualcircuit': frozenset(['cid', 'comments', 'custom_fields', 'description', 'provider_account', 'provider_network', 'status', 'tags', 'tenant', 'type']),
    'circuits.virtualcircuittermination': frozenset(['custom_fields', 'description', 'interface', 'role', 'tags', 'virtual_circuit']),
    'circuits.virtualcircuittype': frozenset(['color', 'custom_fields', 'description', 'name', 'slug', 'tags']),
    'dcim.cable': frozenset(['a_terminations', 'b_terminations', 'color', 'comments', 'custom_fields', 'description', 'label', 'length', 'length_unit', 'status', 'tags', 'tenant', 'type']),
    'dcim.cablepath': frozenset(['is_active', 'is_complete', 'is_split']),
    'dcim.cabletermination': frozenset(['cable', 'cable_end', 'termination_id', 'termination_type']),
    'dcim.consoleport': frozenset(['custom_fields', 'description', 'device', 'label', 'mark_connected', 'module', 'name', 'speed', 'tags', 'type']),
    'dcim.consoleserverport': frozenset(['custom_fields', 'description', 'device', 'label', 'mark_connected', 'module', 'name', 'speed', 'tags', 'type']),
    'dcim.device': frozenset(['airflow', 'asset_tag', 'cluster', 'comments', 'custom_fields', 'description', 'device_type', 'face', 'latitude', 'location', 'longitude', 'name', 'oob_ip', 'platform', 'position', 'primary_ip4', 'primary_ip6', 'rack', 'role', 'serial', 'site', 'status', 'tags', 'tenant', 'vc_position', 'vc_priority', 'virtual_chassis']),
    'dcim.devicebay': frozenset(['custom_fields', 'description', 'device', 'installed_device', 'label', 'name', 'tags']),
    'dcim.devicerole': frozenset(['color', 'custom_fields', 'description', 'name', 'slug', 'tags', 'vm_role']),
    'dcim.devicetype': frozenset(['airflow', 'comments', 'custom_fields', 'default_platform', 'description', 'exclude_from_utilization', 'is_full_depth', 'manufacturer', 'model', 'part_number', 'slug', 'subdevice_role', 'tags', 'u_height', 'weight', 'weight_unit']),
    'dcim.frontport': frozenset(['color', 'custom_fields', 'description', 'device', 'label', 'mark_connected', 'module', 'name', 'rear_port', 'rear_port_position', 'tags', 'type']),
    'dcim.interface': frozenset(['bridge', 'custom_fields', 'description', 'device', 'duplex', 'enabled', 'label', 'lag', 'mark_connected', 'mgmt_only', 'mode', 'module', 'mtu', 'name', 'parent', 'poe_mode', 'poe_type', 'primary_mac_address', 'qinq_svlan', 'rf_channel', 'rf_channel_frequency', 'rf_channel_width', 'rf_role', 'speed', 'tagged_vlans', 'tags', 'tx_power', 'type', 'untagged_vlan', 'vdcs', 'vlan_translation_policy', 'vrf', 'wireless_lans', 'wwn']),
    'dcim.inventoryitem': frozenset(['asset_tag', 'component_id', 'component_type', 'custom_fields', 'description', 'device', 'discovered', 'label', 'manufacturer', 'name', 'parent', 'part_id', 'role', 'serial', 'status', 'tags']),
    'dcim.inventoryitemrole': frozenset(['color', 'custom_fields', 'description', 'name', 'slug', 'tags']),
    'dcim.location': frozenset(['custom_fields', 'description', 'facility', 'name', 'parent', 'site', 'slug', 'status', 'tags', 'tenant']),
    'dcim.macaddress': frozenset(['assigned_object_id', 'assigned_object_type', 'comments', 'custom_fields', 'description', 'mac_address', 'tags']),
    'dcim.manufacturer': frozenset(['custom_fields', 'description', 'name', 'slug', 'tags']),
    'dcim.module': frozenset(['asset_tag', 'comments', 'custom_fields', 'description', 'device', 'module_bay', 'module_type', 'serial', 'status', 'tags']),
    'dcim.modulebay': frozenset(['custom_fields', 'description', 'device', 'installed_module', 'label', 'module', 'name', 'position', 'tags']),
    'dcim.moduletype': frozenset(['airflow', 'comments', 'custom_fields', 'description', 'manufacturer', 'model', 'part_number', 'tags', 'weight', 'weight_unit']),
    'dcim.platform': frozenset(['custom_fields', 'description', 'manufacturer', 'name', 'slug', 'tags']),
    'dcim.powerfeed': frozenset(['amperage', 'comments', 'custom_fields', 'description', 'mark_connected', 'max_utilization', 'name', 'phase', 'power_panel', 'rack', 'status', 'supply', 'tags', 'tenant', 'type', 'voltage']),
    'dcim.poweroutlet': frozenset(['color', 'custom_fields', 'description', 'device', 'feed_leg', 'label', 'mark_connected', 'module', 'name', 'power_port', 'tags', 'type']),
    'dcim.powerpanel': frozenset(['comments', 'custom_fields', 'description', 'location', 'name', 'site', 'tags']),
    'dcim.powerport': frozenset(['allocated_draw', 'custom_fields', 'description', 'device', 'label', 'mark_connected', 'maximum_draw', 'module', 'name', 'tags', 'type']),
    'dcim.rack': frozenset(['airflow', 'asset_tag', 'comments', 'custom_fields', 'desc_units', 'description', 'facility_id', 'form_factor', 'location', 'max_weight', 'mounting_depth', 'name', 'outer_depth', 'outer_unit', 'outer_width', 'rack_type', 'role', 'serial', 'site', 'starting_unit', 'status', 'tags', 'tenant', 'u_height', 'weight', 'weight_unit', 'width']),
    'dcim.rackreservation': frozenset(['comments', 'custom_fields', 'description', 'rack', 'tags', 'tenant', 'units']),
    'dcim.rackrole': frozenset(['color', 'custom_fields', 'description', 'name', 'slug', 'tags']),
    'dcim.racktype': frozenset(['comments', 'custom_fields', 'desc_units', 'description', 'form_factor', 'manufacturer', 'max_weight', 'model', 'mounting_depth', 'outer_depth', 'outer_unit', 'outer_width', 'slug', 'starting_unit', 'tags', 'u_height', 'weight', 'weight_unit', 'width']),
    'dcim.rearport': frozenset(['color', 'custom_fields', 'description', 'device', 'label', 'mark_connected', 'module', 'name', 'positions', 'tags', 'type']),
    'dcim.region': frozenset(['custom_fields', 'description', 'name', 'parent', 'slug', 'tags']),
    'dcim.site': frozenset(['asns', 'comments', 'custom_fields', 'description', 'facility', 'group', 'latitude', 'longitude', 'name', 'physical_address', 'region', 'shipping_address', 'slug', 'status', 'tags', 'tenant', 'time_zone']),
    'dcim.sitegroup': frozenset(['custom_fields', 'description', 'name', 'parent', 'slug', 'tags']),
    'dcim.virtualchassis': frozenset(['comments', 'custom_fields', 'description', 'domain', 'master', 'name', 'tags']),
    'dcim.virtualdevicecontext': frozenset(['comments', 'custom_fields', 'description', 'device', 'identifier', 'name', 'primary_ip4', 'primary_ip6', 'status', 'tags', 'tenant']),
    'extras.tag': frozenset(['color', 'name', 'slug']),
    'ipam.aggregate': frozenset(['comments', 'custom_fields', 'date_added', 'description', 'prefix', 'rir', 'tags', 'tenant']),
    'ipam.asn': frozenset(['asn', 'comments', 'custom_fields', 'description', 'rir', 'tags', 'tenant']),
    'ipam.asnrange': frozenset(['custom_fields', 'description', 'end', 'name', 'rir', 'slug', 'start', 'tags', 'tenant']),
    'ipam.fhrpgroup': frozenset(['auth_key', 'auth_type', 'comments', 'custom_fields', 'description', 'group_id', 'name', 'protocol', 'tags']),
    'ipam.fhrpgroupassignment': frozenset(['group', 'interface_id', 'interface_type', 'priority']),
    'ipam.ipaddress': frozenset(['address', 'assigned_object_id', 'assigned_object_type', 'comments', 'custom_fields', 'description', 'dns_name', 'nat_inside', 'role', 'status', 'tags', 'tenant', 'vrf']),
    'ipam.iprange': frozenset(['comments', 'custom_fields', 'description', 'end_address', 'mark_utilized', 'role', 'start_address', 'status', 'tags', 'tenant', 'vrf']),
    'ipam.prefix': frozenset(['comments', 'custom_fields', 'description', 'is_pool', 'mark_utilized', 'prefix', 'role', 'scope_id', 'scope_type', 'status', 'tags', 'tenant', 'vlan', 'vrf']),
    'ipam.rir': frozenset(['custom_fields', 'description', 'is_private', 'name', 'slug', 'tags']),
    'ipam.role': frozenset(['custom_fields', 'description', 'name', 'slug', 'tags', 'weight']),
    'ipam.routetarget': frozenset(['comments', 'custom_fields', 'description', 'name', 'tags', 'tenant']),
    'ipam.service': frozenset(['comments', 'custom_fields', 'description', 'device', 'ipaddresses', 'name', 'ports', 'protocol', 'tags', 'virtual_machine']),
    'ipam.vlan': frozenset(['comments', 'custom_fields', 'description', 'group', 'name', 'qinq_role', 'qinq_svlan', 'role', 'site', 'status', 'tags', 'tenant', 'vid']),
    'ipam.vlangroup': frozenset(['custom_fields', 'description', 'name', 'scope_id', 'scope_type', 'slug', 'tags', 'vid_ranges']),
    'ipam.vlantranslationpolicy': frozenset(['description', 'name']),
    'ipam.vlantranslationrule': frozenset(['description', 'local_vid', 'policy', 'remote_vid']),
    'ipam.vrf': frozenset(['comments', 'custom_fields', 'description', 'enforce_unique', 'export_targets', 'import_targets', 'name', 'rd', 'tags', 'tenant']),
    'tenancy.contact': frozenset(['address', 'comments', 'custom_fields', 'description', 'email', 'group', 'link', 'name', 'phone', 'tags', 'title']),
    'tenancy.contactassignment': frozenset(['contact', 'custom_fields', 'object_id', 'object_type', 'priority', 'role', 'tags']),
    'tenancy.contactgroup': frozenset(['custom_fields', 'description', 'name', 'parent', 'slug', 'tags']),
    'tenancy.contactrole': frozenset(['custom_fields', 'description', 'name', 'slug', 'tags']),
    'tenancy.tenant': frozenset(['comments', 'custom_fields', 'description', 'group', 'name', 'slug', 'tags']),
    'tenancy.tenantgroup': frozenset(['custom_fields', 'description', 'name', 'parent', 'slug', 'tags']),
    'virtualization.cluster': frozenset(['comments', 'custom_fields', 'description', 'group', 'name', 'scope_id', 'scope_type', 'status', 'tags', 'tenant', 'type']),
    'virtualization.clustergroup': frozenset(['custom_fields', 'description', 'name', 'slug', 'tags']),
    'virtualization.clustertype': frozenset(['custom_fields', 'description', 'name', 'slug', 'tags']),
    'virtualization.virtualdisk': frozenset(['custom_fields', 'description', 'name', 'size', 'tags', 'virtual_machine']),
    'virtualization.virtualmachine': frozenset(['cluster', 'comments', 'custom_fields', 'description', 'device', 'disk', 'memory', 'name', 'platform', 'primary_ip4', 'primary_ip6', 'role', 'serial', 'site', 'status', 'tags', 'tenant', 'vcpus']),
    'virtualization.vminterface': frozenset(['bridge', 'custom_fields', 'description', 'enabled', 'mode', 'mtu', 'name', 'parent', 'primary_mac_address', 'qinq_svlan', 'tagged_vlans', 'tags', 'untagged_vlan', 'virtual_machine', 'vlan_translation_policy', 'vrf']),
    'vpn.ikepolicy': frozenset(['comments', 'custom_fields', 'description', 'mode', 'name', 'preshared_key', 'proposals', 'tags', 'version']),
    'vpn.ikeproposal': frozenset(['authentication_algorithm', 'authentication_method', 'comments', 'custom_fields', 'description', 'encryption_algorithm', 'group', 'name', 'sa_lifetime', 'tags']),
    'vpn.ipsecpolicy': frozenset(['comments', 'custom_fields', 'description', 'name', 'pfs_group', 'proposals', 'tags']),
    'vpn.ipsecprofile': frozenset(['comments', 'custom_fields', 'description', 'ike_policy', 'ipsec_policy', 'mode', 'name', 'tags']),
    'vpn.ipsecproposal': frozenset(['authentication_algorithm', 'comments', 'custom_fields', 'description', 'encryption_algorithm', 'name', 'sa_lifetime_data', 'sa_lifetime_seconds', 'tags']),
    'vpn.l2vpn': frozenset(['comments', 'custom_fields', 'description', 'export_targets', 'identifier', 'import_targets', 'name', 'slug', 'tags', 'tenant', 'type']),
    'vpn.l2vpntermination': frozenset(['assigned_object_id', 'assigned_object_type', 'custom_fields', 'l2vpn', 'tags']),
    'vpn.tunnel': frozenset(['comments', 'custom_fields', 'description', 'encapsulation', 'group', 'ipsec_profile', 'name', 'status', 'tags', 'tenant', 'tunnel_id']),
    'vpn.tunnelgroup': frozenset(['custom_fields', 'description', 'name', 'slug', 'tags']),
    'vpn.tunneltermination': frozenset(['custom_fields', 'outside_ip', 'role', 'tags', 'termination_id', 'termination_type', 'tunnel']),
    'wireless.wirelesslan': frozenset(['auth_cipher', 'auth_psk', 'auth_type', 'comments', 'custom_fields', 'description', 'group', 'scope_id', 'scope_type', 'ssid', 'status', 'tags', 'tenant', 'vlan']),
    'wireless.wirelesslangroup': frozenset(['custom_fields', 'description', 'name', 'parent', 'slug', 'tags']),
    'wireless.wirelesslink': frozenset(['auth_cipher', 'auth_psk', 'auth_type', 'comments', 'custom_fields', 'description', 'distance', 'distance_unit', 'interface_a', 'interface_b', 'ssid', 'status', 'tags', 'tenant']),
}

def legal_fields(object_type: str|Type[models.Model]) -> frozenset[str]:
    if not isinstance(object_type, str):
        object_type = get_object_type(object_type)
    return _LEGAL_FIELDS.get(object_type, frozenset())

_OBJECT_TYPE_PRIMARY_VALUE_FIELD_MAP = {
    'ipam.asn': 'asn',
    'dcim.devicetype': 'model',
    'circuits.circuit': 'cid',
    'ipam.ipaddress': 'address',
    'dcim.macaddress': 'mac_address',
    'dcim.moduletype': 'model',
    'ipam.prefix': 'prefix',
    'dcim.racktype': 'model',
    'circuits.virtualcircuit': 'cid',
    'wireless.wirelesslan': 'ssid',
}

def get_primary_value(data: dict, object_type: str) -> str|None:
    field = _OBJECT_TYPE_PRIMARY_VALUE_FIELD_MAP.get(object_type, 'name')
    return data.get(field)


def transform_timestamp_to_date_only(value: str) -> str:
    return datetime.datetime.fromisoformat(value).strftime('%Y-%m-%d')

def transform_float_to_decimal(value: float) -> decimal.Decimal:
    try:
        return decimal.Decimal(str(value))
    except decimal.InvalidOperation:
        raise ValueError(f'Invalid decimal value: {value}')

def int_from_int64string(value: str) -> int:
    return int(value)

def ip_network_defaulting(value: str) -> str:
    try:
        return str(netaddr.IPNetwork(value))
    except netaddr.AddrFormatError:
        raise ValueError(f'Invalid IP network value: {value}')

def collect_integer_pairs(value: list[int]) -> list[tuple[int, int]]:
    if len(value) % 2 != 0:
        raise ValueError('Array must have an even number of elements')
    return sorted([(value[i], value[i+1]) for i in range(0, len(value), 2)])

def for_all(transform):
    def wrapper(value):
        if isinstance(value, list):
            return [transform(v) for v in value]
        return transform(value)
    return wrapper

_FORMAT_TRANSFORMATIONS = {
    'circuits.circuit': {
        'commit_rate': int_from_int64string,
        'distance': transform_float_to_decimal,
        'install_date': transform_timestamp_to_date_only,
        'termination_date': transform_timestamp_to_date_only,
    },
    'circuits.circuittermination': {
        'port_speed': int_from_int64string,
        'upstream_speed': int_from_int64string,
    },
    'dcim.cable': {
        'length': transform_float_to_decimal,
    },
    'dcim.consoleport': {
        'speed': int_from_int64string,
    },
    'dcim.consoleserverport': {
        'speed': int_from_int64string,
    },
    'dcim.device': {
        'latitude': transform_float_to_decimal,
        'longitude': transform_float_to_decimal,
        'position': transform_float_to_decimal,
        'vc_position': int_from_int64string,
        'vc_priority': int_from_int64string,
    },
    'dcim.devicetype': {
        'u_height': transform_float_to_decimal,
        'weight': transform_float_to_decimal,
    },
    'dcim.frontport': {
        'rear_port_position': int_from_int64string,
    },
    'dcim.interface': {
        'mtu': int_from_int64string,
        'rf_channel_frequency': transform_float_to_decimal,
        'rf_channel_width': transform_float_to_decimal,
        'speed': int_from_int64string,
        'tx_power': int_from_int64string,
    },
    'dcim.moduletype': {
        'weight': transform_float_to_decimal,
    },
    'dcim.powerfeed': {
        'amperage': int_from_int64string,
        'max_utilization': int_from_int64string,
        'voltage': int_from_int64string,
    },
    'dcim.powerport': {
        'allocated_draw': int_from_int64string,
        'maximum_draw': int_from_int64string,
    },
    'dcim.rack': {
        'max_weight': int_from_int64string,
        'mounting_depth': int_from_int64string,
        'outer_depth': int_from_int64string,
        'outer_width': int_from_int64string,
        'starting_unit': int_from_int64string,
        'u_height': int_from_int64string,
        'weight': transform_float_to_decimal,
        'width': int_from_int64string,
    },
    'dcim.rackreservation': {
        'units': for_all(int_from_int64string),
    },
    'dcim.racktype': {
        'max_weight': int_from_int64string,
        'mounting_depth': int_from_int64string,
        'outer_depth': int_from_int64string,
        'outer_width': int_from_int64string,
        'starting_unit': int_from_int64string,
        'u_height': int_from_int64string,
        'weight': transform_float_to_decimal,
        'width': int_from_int64string,
    },
    'dcim.rearport': {
        'positions': int_from_int64string,
    },
    'dcim.site': {
        'latitude': transform_float_to_decimal,
        'longitude': transform_float_to_decimal,
    },
    'dcim.virtualdevicecontext': {
        'identifier': int_from_int64string,
    },
    'ipam.aggregate': {
        'date_added': transform_timestamp_to_date_only,
        'prefix': ip_network_defaulting,
    },
    'ipam.asn': {
        'asn': int_from_int64string,
    },
    'ipam.asnrange': {
        'end': int_from_int64string,
        'start': int_from_int64string,
    },
    'ipam.fhrpgroup': {
        'group_id': int_from_int64string,
    },
    'ipam.fhrpgroupassignment': {
        'priority': int_from_int64string,
    },
    'ipam.ipaddress': {
        'address': ip_network_defaulting,
    },
    'ipam.iprange': {
        'end_address': ip_network_defaulting,
        'start_address': ip_network_defaulting,
    },
    'ipam.prefix': {
        'prefix': ip_network_defaulting,
    },
    'ipam.role': {
        'weight': int_from_int64string,
    },
    'ipam.service': {
        'ports': for_all(int_from_int64string),
    },
    'ipam.vlan': {
        'vid': int_from_int64string,
    },
    'ipam.vlangroup': {
        'vid_ranges': collect_integer_pairs,
    },
    'ipam.vlantranslationrule': {
        'local_vid': int_from_int64string,
        'remote_vid': int_from_int64string,
    },
    'virtualization.virtualdisk': {
        'size': int_from_int64string,
    },
    'virtualization.virtualmachine': {
        'disk': int_from_int64string,
        'memory': int_from_int64string,
        'vcpus': transform_float_to_decimal,
    },
    'virtualization.vminterface': {
        'mtu': int_from_int64string,
    },
    'vpn.ikepolicy': {
        'version': int_from_int64string,
    },
    'vpn.ikeproposal': {
        'group': int_from_int64string,
        'sa_lifetime': int_from_int64string,
    },
    'vpn.ipsecpolicy': {
        'pfs_group': int_from_int64string,
    },
    'vpn.ipsecproposal': {
        'sa_lifetime_data': int_from_int64string,
        'sa_lifetime_seconds': int_from_int64string,
    },
    'vpn.l2vpn': {
        'identifier': int_from_int64string,
    },
    'vpn.tunnel': {
        'tunnel_id': int_from_int64string,
    },
    'wireless.wirelesslink': {
        'distance': transform_float_to_decimal,
    },
}

def apply_format_transformations(data: dict, object_type: str):
    for key, transform in _FORMAT_TRANSFORMATIONS.get(object_type, {}).items():
        val = data.get(key, None)
        if val is None:
            continue
        try:
            data[key] = transform(val)
        except ValidationError:
            raise
        except ValueError as e:
            sanitized_object_type = object_type.replace('\n', '').replace('\r', '')
            sanitized_val = str(val).replace('\n', '').replace('\r', '')
            logger.error(f"Error processing field {key} in {sanitized_object_type} with value {sanitized_val}: {e}")
            raise ValidationError(f"Invalid value for field {key} in {sanitized_object_type}.")
        except Exception as e:
            raise ValidationError(f'Invalid value for field {key} in {object_type}')


#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin."""

from netbox.plugins import PluginConfig

from .version import version_semver


class NetBoxDiodePluginConfig(PluginConfig):
    """NetBox Diode plugin configuration."""

    name = "netbox_diode_plugin"
    verbose_name = "NetBox Labs, Diode Plugin"
    description = "Diode plugin for NetBox."
    version = version_semver()
    base_url = "diode"
    min_version = "4.2.3"
    default_settings = {
        # Default Diode gRPC target for communication with Diode server
        "diode_target": "grpc://localhost:8080/diode",

        # Default username associated with changes applied via plugin
        "diode_username": "diode",
    }


config = NetBoxDiodePluginConfig



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode Netbox Plugin - URLs."""

from django.urls import path

from . import views

urlpatterns = (
    path("settings/", views.SettingsView.as_view(), name="settings"),
    path("settings/edit/", views.SettingsEditView.as_view(), name="settings_edit"),
)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Navigation."""

from netbox.plugins import PluginMenu, PluginMenuItem

settings = {
    "link": "plugins:netbox_diode_plugin:settings",
    "link_text": "Settings",
    "staff_only": True,
}


menu = PluginMenu(
    label="Diode",
    groups=(
        (
            "Diode",
            (
                PluginMenuItem(**settings),
            ),
        ),
    ),
    icon_class="mdi mdi-upload",
)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Views."""
from django.conf import settings as netbox_settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.utils.http import url_has_allowed_host_and_scheme
from django.views.generic import View
from netbox.plugins import get_plugin_config
from netbox.views import generic
from utilities.views import register_model_view

from netbox_diode_plugin.forms import SettingsForm
from netbox_diode_plugin.models import Setting

User = get_user_model()


def redirect_to_login(request):
    """Redirect to login view."""
    redirect_url = netbox_settings.LOGIN_URL
    target = request.path

    if target and url_has_allowed_host_and_scheme(target, allowed_hosts=None):
        redirect_url = f"{netbox_settings.LOGIN_URL}?next={target}"

    return HttpResponseRedirect(redirect_url)


class SettingsView(View):
    """Settings view."""

    def get(self, request):
        """Render settings template."""
        if not request.user.is_authenticated or not request.user.is_staff:
            return redirect_to_login(request)

        diode_target_override = get_plugin_config(
            "netbox_diode_plugin", "diode_target_override"
        )

        try:
            settings = Setting.objects.get()
        except Setting.DoesNotExist:
            default_diode_target = get_plugin_config(
                "netbox_diode_plugin", "diode_target"
            )
            settings = Setting.objects.create(
                diode_target=diode_target_override or default_diode_target
            )

        diode_target = diode_target_override or settings.diode_target

        context = {
            "diode_target": diode_target,
            "is_diode_target_overridden": diode_target_override is not None,
        }

        return render(request, "diode/settings.html", context)


@register_model_view(Setting, "edit")
class SettingsEditView(generic.ObjectEditView):
    """Settings edit view."""

    queryset = Setting.objects
    form = SettingsForm
    template_name = "diode/settings_edit.html"
    default_return_url = "plugins:netbox_diode_plugin:settings"

    def get(self, request, *args, **kwargs):
        """GET request handler."""
        if not request.user.is_authenticated or not request.user.is_staff:
            return redirect_to_login(request)

        diode_target_override = get_plugin_config(
            "netbox_diode_plugin", "diode_target_override"
        )
        if diode_target_override:
            messages.error(
                request,
                "The Diode target is not allowed to be modified.",
            )
            return redirect("plugins:netbox_diode_plugin:settings")

        settings = Setting.objects.get()
        kwargs["pk"] = settings.pk

        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """POST request handler."""
        if not request.user.is_authenticated or not request.user.is_staff:
            return redirect_to_login(request)

        diode_target_override = get_plugin_config(
            "netbox_diode_plugin", "diode_target_override"
        )
        if diode_target_override:
            messages.error(
                request,
                "The Diode target is not allowed to be modified.",
            )
            return redirect("plugins:netbox_diode_plugin:settings")

        settings = Setting.objects.get()
        kwargs["pk"] = settings.pk

        return super().post(request, *args, **kwargs)
